#!/usr/bin/env python3
"""Infer forwarding-based parameter bundles and propagate them across calls.

This script performs a two-stage analysis:
  1) Local grouping: within a function, parameters used *only* as direct
     call arguments are grouped by identical forwarding signatures.
  2) Propagation: if a function f calls g, and g has local bundles, then
     f's parameters passed into g's bundled positions are linked as a
     candidate bundle. This is iterated to a fixed point.

The goal is to surface "dataflow grammar" candidates for config dataclasses.

It can also emit a DOT graph (see --dot) so downstream tooling can render
bundle candidates as a dependency graph.
"""
from __future__ import annotations

import argparse
import ast
import os
from collections import defaultdict, deque
from dataclasses import dataclass
from pathlib import Path
from typing import Iterable, Iterator
import re


@dataclass
class ParamUse:
    direct_forward: set[tuple[str, str]]
    non_forward: bool


@dataclass(frozen=True)
class CallArgs:
    callee: str
    pos_map: dict[str, str]
    kw_map: dict[str, str]
    const_pos: dict[str, str]
    const_kw: dict[str, str]
    non_const_pos: set[str]
    non_const_kw: set[str]
    is_test: bool


class ParentAnnotator(ast.NodeVisitor):
    def __init__(self) -> None:
        self.parents: dict[ast.AST, ast.AST] = {}

    def generic_visit(self, node: ast.AST) -> None:
        for child in ast.iter_child_nodes(node):
            self.parents[child] = node
            self.visit(child)


def _call_context(node: ast.AST, parents: dict[ast.AST, ast.AST]) -> tuple[ast.Call | None, bool]:
    child = node
    parent = parents.get(child)
    while parent is not None:
        if isinstance(parent, ast.Call):
            if child in parent.args:
                return parent, True
            for kw in parent.keywords:
                if child is kw or child is kw.value:
                    return parent, True
            return parent, False
        child = parent
        parent = parents.get(child)
    return None, False


def _callee_name(call: ast.Call) -> str:
    try:
        return ast.unparse(call.func)
    except Exception:
        return "<call>"


def _iter_paths(paths: Iterable[str]) -> list[Path]:
    out: list[Path] = []
    for p in paths:
        path = Path(p)
        if path.is_dir():
            out.extend(sorted(path.rglob("*.py")))
        else:
            out.append(path)
    return out


def _collect_functions(tree: ast.AST):
    funcs = []
    for node in ast.walk(tree):
        if isinstance(node, (ast.FunctionDef, ast.AsyncFunctionDef)):
            funcs.append(node)
    return funcs


def _param_names(fn: ast.FunctionDef | ast.AsyncFunctionDef) -> list[str]:
    args = (
        fn.args.posonlyargs + fn.args.args + fn.args.kwonlyargs
    )
    names = [a.arg for a in args]
    if fn.args.vararg:
        names.append(fn.args.vararg.arg)
    if fn.args.kwarg:
        names.append(fn.args.kwarg.arg)
    if names and names[0] in {"self", "cls"}:
        names = names[1:]
    return names


def _param_annotations(fn: ast.FunctionDef | ast.AsyncFunctionDef) -> dict[str, str | None]:
    args = fn.args.posonlyargs + fn.args.args + fn.args.kwonlyargs
    names = [a.arg for a in args]
    annots: dict[str, str | None] = {}
    for name, arg in zip(names, args):
        if arg.annotation is None:
            annots[name] = None
        else:
            try:
                annots[name] = ast.unparse(arg.annotation)
            except Exception:
                annots[name] = None
    if fn.args.vararg:
        annots[fn.args.vararg.arg] = None
    if fn.args.kwarg:
        annots[fn.args.kwarg.arg] = None
    if names and names[0] in {"self", "cls"}:
        annots.pop(names[0], None)
    return annots


def _const_repr(node: ast.AST) -> str | None:
    if isinstance(node, ast.Constant):
        return repr(node.value)
    if isinstance(node, ast.UnaryOp) and isinstance(
        node.op, (ast.USub, ast.UAdd)
    ) and isinstance(node.operand, ast.Constant):
        try:
            return ast.unparse(node)
        except Exception:
            return None
    if isinstance(node, ast.Attribute):
        if node.attr.isupper():
            try:
                return ast.unparse(node)
            except Exception:
                return None
        return None
    return None


def _is_test_path(path: Path) -> bool:
    if "tests" in path.parts:
        return True
    return path.name.startswith("test_")


def _analyze_function(
    fn: ast.FunctionDef | ast.AsyncFunctionDef,
    parents: dict[ast.AST, ast.AST],
    *,
    is_test: bool,
):
    params = _param_names(fn)
    use_map = {p: ParamUse(set(), False) for p in params}
    call_args: list[CallArgs] = []

    class UseVisitor(ast.NodeVisitor):
        def visit_Call(self, node: ast.Call) -> None:
            callee = _callee_name(node)
            pos_map = {}
            kw_map = {}
            const_pos: dict[str, str] = {}
            const_kw: dict[str, str] = {}
            non_const_pos: set[str] = set()
            non_const_kw: set[str] = set()
            for idx, arg in enumerate(node.args):
                const = _const_repr(arg)
                if const is not None:
                    const_pos[str(idx)] = const
                    continue
                if isinstance(arg, ast.Name) and arg.id in use_map:
                    pos_map[str(idx)] = arg.id
                else:
                    non_const_pos.add(str(idx))
            for kw in node.keywords:
                if kw.arg is None:
                    continue
                const = _const_repr(kw.value)
                if const is not None:
                    const_kw[kw.arg] = const
                    continue
                if isinstance(kw.value, ast.Name) and kw.value.id in use_map:
                    kw_map[kw.arg] = kw.value.id
                else:
                    non_const_kw.add(kw.arg)
            call_args.append(
                CallArgs(
                    callee=callee,
                    pos_map=pos_map,
                    kw_map=kw_map,
                    const_pos=const_pos,
                    const_kw=const_kw,
                    non_const_pos=non_const_pos,
                    non_const_kw=non_const_kw,
                    is_test=is_test,
                )
            )
            self.generic_visit(node)

        def visit_Name(self, node: ast.Name) -> None:
            if not isinstance(node.ctx, ast.Load):
                return
            if node.id not in use_map:
                return
            call, direct = _call_context(node, parents)
            if call is None or not direct:
                use_map[node.id].non_forward = True
                return
            callee = _callee_name(call)
            # Determine arg slot.
            slot = None
            for idx, arg in enumerate(call.args):
                if arg is node:
                    slot = f"arg[{idx}]"
                    break
            if slot is None:
                for kw in call.keywords:
                    if kw.value is node and kw.arg is not None:
                        slot = f"kw[{kw.arg}]"
                        break
            if slot is None:
                slot = "arg[?]"
            use_map[node.id].direct_forward.add((callee, slot))

    UseVisitor().visit(fn)
    return use_map, call_args


def _group_by_signature(use_map: dict[str, ParamUse]) -> list[set[str]]:
    sig_map: dict[tuple[tuple[str, str], ...], list[str]] = defaultdict(list)
    for name, info in use_map.items():
        if info.non_forward:
            continue
        sig = tuple(sorted(info.direct_forward))
        sig_map[sig].append(name)
    groups = [set(names) for names in sig_map.values() if len(names) > 1]
    return groups


def _union_groups(groups: list[set[str]]) -> list[set[str]]:
    changed = True
    while changed:
        changed = False
        out = []
        while groups:
            base = groups.pop()
            merged = True
            while merged:
                merged = False
                for i, other in enumerate(groups):
                    if base & other:
                        base |= other
                        groups.pop(i)
                        merged = True
                        changed = True
                        break
            out.append(base)
        groups = out
    return groups


def _propagate_groups(
    call_args: list[CallArgs],
    callee_groups: dict[str, list[set[str]]],
    callee_param_orders: dict[str, list[str]],
) -> list[set[str]]:
    groups: list[set[str]] = []
    for call in call_args:
        if call.callee not in callee_groups:
            continue
        callee_params = callee_param_orders[call.callee]
        # Build mapping from callee param to caller param.
        callee_to_caller: dict[str, str] = {}
        for idx, pname in enumerate(callee_params):
            key = str(idx)
            if key in call.pos_map:
                callee_to_caller[pname] = call.pos_map[key]
        for kw, caller_name in call.kw_map.items():
            callee_to_caller[kw] = caller_name
        for group in callee_groups[call.callee]:
            mapped = {callee_to_caller.get(p) for p in group}
            mapped.discard(None)
            if len(mapped) > 1:
                groups.append(set(mapped))
    return groups


def analyze_file(path: Path, recursive: bool = True) -> dict[str, list[set[str]]]:
    tree = ast.parse(path.read_text())
    parent = ParentAnnotator()
    parent.visit(tree)
    parents = parent.parents
    is_test = _is_test_path(path)

    funcs = _collect_functions(tree)
    fn_param_orders = {f.name: _param_names(f) for f in funcs}
    fn_use = {}
    fn_calls = {}
    for f in funcs:
        use_map, call_args = _analyze_function(f, parents, is_test=is_test)
        fn_use[f.name] = use_map
        fn_calls[f.name] = call_args

    groups_by_fn = {fn: _group_by_signature(use_map) for fn, use_map in fn_use.items()}

    if not recursive:
        return groups_by_fn

    changed = True
    while changed:
        changed = False
        for fn in fn_use:
            propagated = _propagate_groups(
                fn_calls[fn],
                groups_by_fn,
                fn_param_orders,
            )
            if not propagated:
                continue
            combined = _union_groups(groups_by_fn.get(fn, []) + propagated)
            if combined != groups_by_fn.get(fn, []):
                groups_by_fn[fn] = combined
                changed = True
    return groups_by_fn


def _callee_key(name: str) -> str:
    if not name:
        return name
    return name.split(".")[-1]


def _is_broad_type(annot: str | None) -> bool:
    if annot is None:
        return True
    base = annot.replace("typing.", "")
    return base in {"Any", "object"}


@dataclass
class FunctionInfo:
    name: str
    qual: str
    path: Path
    params: list[str]
    annots: dict[str, str | None]
    calls: list[CallArgs]


def _module_name(path: Path) -> str:
    rel = path.with_suffix("")
    return ".".join(rel.parts)


def _build_function_index(paths: list[Path]) -> tuple[dict[str, list[FunctionInfo]], dict[str, FunctionInfo]]:
    by_name: dict[str, list[FunctionInfo]] = defaultdict(list)
    by_qual: dict[str, FunctionInfo] = {}
    for path in paths:
        try:
            tree = ast.parse(path.read_text())
        except Exception:
            continue
        funcs = _collect_functions(tree)
        if not funcs:
            continue
        parents = ParentAnnotator()
        parents.visit(tree)
        parent_map = parents.parents
        module = _module_name(path)
        for fn in funcs:
            _, call_args = _analyze_function(
                fn, parent_map, is_test=_is_test_path(path)
            )
            info = FunctionInfo(
                name=fn.name,
                qual=f"{module}.{fn.name}",
                path=path,
                params=_param_names(fn),
                annots=_param_annotations(fn),
                calls=call_args,
            )
            by_name[fn.name].append(info)
            by_qual[info.qual] = info
    return by_name, by_qual


def _resolve_callee(
    callee_name: str,
    caller: FunctionInfo,
    by_name: dict[str, list[FunctionInfo]],
    by_qual: dict[str, FunctionInfo],
) -> FunctionInfo | None:
    if not callee_name:
        return None
    # Exact qualified name match.
    if callee_name in by_qual:
        return by_qual[callee_name]
    # If call uses module.func, try match by module suffix.
    if "." in callee_name:
        parts = callee_name.split(".")
        func = parts[-1]
        module = ".".join(parts[:-1])
        candidates = [
            info
            for info in by_name.get(func, [])
            if info.qual.endswith(f"{module}.{func}")
        ]
        if len(candidates) == 1:
            return candidates[0]
        # If caller's module matches a candidate, prefer it.
        same_module = [
            info
            for info in candidates
            if info.path == caller.path
        ]
        if len(same_module) == 1:
            return same_module[0]
        return None
    # Fallback: unique function name across repo.
    candidates = by_name.get(callee_name, [])
    if len(candidates) == 1:
        return candidates[0]
    # Prefer same-module definition when ambiguous.
    same_module = [info for info in candidates if info.path == caller.path]
    if len(same_module) == 1:
        return same_module[0]
    # If callee is self.foo/cls.foo, prefer same-module foo.
    if callee_name.startswith(("self.", "cls.")):
        func = callee_name.split(".")[-1]
        same_module = [
            info for info in by_name.get(func, []) if info.path == caller.path
        ]
        if len(same_module) == 1:
            return same_module[0]
    return None


def analyze_type_flow_repo(paths: list[Path]) -> tuple[list[str], list[str]]:
    """Repo-wide fixed-point pass for downstream type tightening."""
    by_name, by_qual = _build_function_index(paths)
    inferred: dict[str, dict[str, str | None]] = {}
    for infos in by_name.values():
        for info in infos:
            inferred[info.qual] = dict(info.annots)

    def _get_annot(info: FunctionInfo, param: str) -> str | None:
        return inferred.get(info.qual, {}).get(param)

    suggestions: set[str] = set()
    ambiguities: set[str] = set()
    changed = True
    while changed:
        changed = False
        for infos in by_name.values():
            for info in infos:
                downstream: dict[str, set[str]] = defaultdict(set)
                for call in info.calls:
                    callee = _resolve_callee(call.callee, info, by_name, by_qual)
                    if callee is None:
                        continue
                    callee_params = callee.params
                    for pos_idx, param in call.pos_map.items():
                        try:
                            idx = int(pos_idx)
                        except ValueError:
                            continue
                        if idx >= len(callee_params):
                            continue
                        callee_param = callee_params[idx]
                        annot = _get_annot(callee, callee_param)
                        if annot:
                            downstream[param].add(annot)
                    for kw_name, param in call.kw_map.items():
                        annot = _get_annot(callee, kw_name)
                        if annot:
                            downstream[param].add(annot)
                for param, annots in downstream.items():
                    if not annots:
                        continue
                    if len(annots) > 1:
                        ambiguities.add(
                            f"{info.path.name}:{info.name}.{param} downstream types conflict: {sorted(annots)}"
                        )
                        continue
                    downstream_annot = next(iter(annots))
                    current = _get_annot(info, param)
                    if _is_broad_type(current) and downstream_annot:
                        if inferred[info.qual].get(param) != downstream_annot:
                            inferred[info.qual][param] = downstream_annot
                            changed = True
                        suggestions.add(
                            f"{info.path.name}:{info.name}.{param} can tighten to {downstream_annot}"
                        )
    return sorted(suggestions), sorted(ambiguities)


def analyze_constant_flow_repo(paths: list[Path]) -> list[str]:
    """Detect parameters that only receive a single constant value (non-test)."""
    by_name, by_qual = _build_function_index(paths)
    const_values: dict[tuple[str, str], set[str]] = defaultdict(set)
    non_const: dict[tuple[str, str], bool] = defaultdict(bool)
    call_counts: dict[tuple[str, str], int] = defaultdict(int)

    for infos in by_name.values():
        for info in infos:
            for call in info.calls:
                if call.is_test:
                    continue
                callee = _resolve_callee(call.callee, info, by_name, by_qual)
                if callee is None:
                    continue
                callee_params = callee.params

                for idx_str, value in call.const_pos.items():
                    try:
                        idx = int(idx_str)
                    except ValueError:
                        continue
                    if idx >= len(callee_params):
                        continue
                    key = (callee.qual, callee_params[idx])
                    const_values[key].add(value)
                    call_counts[key] += 1
                for idx_str in call.pos_map:
                    try:
                        idx = int(idx_str)
                    except ValueError:
                        continue
                    if idx >= len(callee_params):
                        continue
                    key = (callee.qual, callee_params[idx])
                    non_const[key] = True
                    call_counts[key] += 1
                for idx_str in call.non_const_pos:
                    try:
                        idx = int(idx_str)
                    except ValueError:
                        continue
                    if idx >= len(callee_params):
                        continue
                    key = (callee.qual, callee_params[idx])
                    non_const[key] = True
                    call_counts[key] += 1

                for kw, value in call.const_kw.items():
                    if kw not in callee_params:
                        continue
                    key = (callee.qual, kw)
                    const_values[key].add(value)
                    call_counts[key] += 1
                for kw in call.kw_map:
                    if kw not in callee_params:
                        continue
                    key = (callee.qual, kw)
                    non_const[key] = True
                    call_counts[key] += 1
                for kw in call.non_const_kw:
                    if kw not in callee_params:
                        continue
                    key = (callee.qual, kw)
                    non_const[key] = True
                    call_counts[key] += 1

    smells: list[str] = []
    for key, values in const_values.items():
        if non_const.get(key):
            continue
        if not values:
            continue
        if len(values) == 1:
            qual, param = key
            info = by_qual.get(qual)
            path_name = info.path.name if info is not None else qual
            count = call_counts.get(key, 0)
            smells.append(
                f"{path_name}:{qual.split('.')[-1]}.{param} only observed constant {next(iter(values))} across {count} non-test call(s)"
            )
    return sorted(smells)


def _iter_config_fields(path: Path) -> dict[str, set[str]]:
    """Best-effort extraction of @dataclass config bundles (fields ending with _fn)."""
    try:
        tree = ast.parse(path.read_text())
    except Exception:
        return {}
    bundles: dict[str, set[str]] = {}
    for node in ast.walk(tree):
        if not isinstance(node, ast.ClassDef):
            continue
        decorators = {getattr(d, "id", None) for d in node.decorator_list}
        if "dataclass" not in decorators and not node.name.endswith("Config"):
            continue
        fields: set[str] = set()
        for stmt in node.body:
            if isinstance(stmt, ast.AnnAssign) and isinstance(stmt.target, ast.Name):
                name = stmt.target.id
                if name.endswith("_fn"):
                    fields.add(name)
            elif isinstance(stmt, ast.Assign):
                for target in stmt.targets:
                    if isinstance(target, ast.Name) and target.id.endswith("_fn"):
                        fields.add(target.id)
        if fields:
            bundles[node.name] = fields
    return bundles


_BUNDLE_MARKER = re.compile(r"dataflow-bundle:\s*(.*)")


def _iter_documented_bundles(path: Path) -> set[tuple[str, ...]]:
    """Return bundles documented via '# dataflow-bundle: a, b' markers."""
    bundles: set[tuple[str, ...]] = set()
    try:
        text = path.read_text()
    except Exception:
        return bundles
    for line in text.splitlines():
        match = _BUNDLE_MARKER.search(line)
        if not match:
            continue
        payload = match.group(1)
        if not payload:
            continue
        parts = [p.strip() for p in re.split(r"[,\s]+", payload) if p.strip()]
        if len(parts) < 2:
            continue
        bundles.add(tuple(sorted(parts)))
    return bundles


def _iter_dataclass_call_bundles(path: Path) -> set[tuple[str, ...]]:
    """Return bundles promoted via local @dataclass constructor calls."""
    bundles: set[tuple[str, ...]] = set()
    try:
        tree = ast.parse(path.read_text())
    except Exception:
        return bundles
    dataclass_names: set[str] = set()
    for node in ast.walk(tree):
        if not isinstance(node, ast.ClassDef):
            continue
        decorators = {
            ast.unparse(dec) if hasattr(ast, "unparse") else ""
            for dec in node.decorator_list
        }
        if any("dataclass" in dec for dec in decorators):
            dataclass_names.add(node.name)
    if not dataclass_names:
        return bundles

    def _callee_name(call: ast.Call) -> str | None:
        if isinstance(call.func, ast.Name):
            return call.func.id
        if isinstance(call.func, ast.Attribute):
            return call.func.attr
        return None

    for node in ast.walk(tree):
        if not isinstance(node, ast.Call):
            continue
        callee = _callee_name(node)
        if callee not in dataclass_names:
            continue
        names: list[str] = []
        ok = True
        for arg in node.args:
            if isinstance(arg, ast.Name):
                names.append(arg.id)
            else:
                ok = False
                break
        if not ok:
            continue
        for kw in node.keywords:
            if kw.arg is None:
                ok = False
                break
            if isinstance(kw.value, ast.Name):
                names.append(kw.value.id)
            else:
                ok = False
                break
        if not ok or len(names) < 2:
            continue
        bundles.add(tuple(sorted(names)))
    return bundles


def _emit_dot(groups_by_path: dict[Path, dict[str, list[set[str]]]]) -> str:
    lines = [
        "digraph dataflow_grammar {",
        "  rankdir=LR;",
        "  node [fontsize=10];",
    ]
    for path, groups in groups_by_path.items():
        file_id = str(path).replace("/", "_").replace(".", "_")
        lines.append(f"  subgraph cluster_{file_id} {{")
        lines.append(f"    label=\"{path}\";")
        for fn, bundles in groups.items():
            if not bundles:
                continue
            fn_id = f"fn_{file_id}_{fn}"
            lines.append(f"    {fn_id} [shape=box,label=\"{fn}\"];")
            for idx, bundle in enumerate(bundles):
                bundle_id = f"b_{file_id}_{fn}_{idx}"
                label = ", ".join(sorted(bundle))
                lines.append(
                    f"    {bundle_id} [shape=ellipse,label=\"{label}\"];"
                )
                lines.append(f"    {fn_id} -> {bundle_id};")
        lines.append("  }")
    lines.append("}")
    return "\n".join(lines)


def _component_graph(groups_by_path: dict[Path, dict[str, list[set[str]]]]):
    nodes: dict[str, dict[str, str]] = {}
    adj: dict[str, set[str]] = defaultdict(set)
    bundle_map: dict[str, set[str]] = {}
    for path, groups in groups_by_path.items():
        file_id = str(path)
        for fn, bundles in groups.items():
            if not bundles:
                continue
            fn_id = f"fn::{file_id}::{fn}"
            nodes[fn_id] = {"kind": "fn", "label": f"{path.name}:{fn}"}
            for idx, bundle in enumerate(bundles):
                bundle_id = f"b::{file_id}::{fn}::{idx}"
                nodes[bundle_id] = {
                    "kind": "bundle",
                    "label": ", ".join(sorted(bundle)),
                }
                bundle_map[bundle_id] = bundle
                adj[fn_id].add(bundle_id)
                adj[bundle_id].add(fn_id)
    return nodes, adj, bundle_map


def _connected_components(nodes: dict[str, dict[str, str]], adj: dict[str, set[str]]) -> list[list[str]]:
    seen: set[str] = set()
    comps: list[list[str]] = []
    for node in nodes:
        if node in seen:
            continue
        q: deque[str] = deque([node])
        seen.add(node)
        comp: list[str] = []
        while q:
            curr = q.popleft()
            comp.append(curr)
            for nxt in adj.get(curr, ()):
                if nxt not in seen:
                    seen.add(nxt)
                    q.append(nxt)
        comps.append(sorted(comp))
    return comps


def _render_mermaid_component(
    nodes: dict[str, dict[str, str]],
    bundle_map: dict[str, set[str]],
    adj: dict[str, set[str]],
    component: list[str],
    config_bundles_by_path: dict[Path, dict[str, set[str]]],
    documented_bundles_by_path: dict[Path, set[tuple[str, ...]]],
) -> tuple[str, str]:
    lines = ["```mermaid", "flowchart LR"]
    fn_nodes = [n for n in component if nodes[n]["kind"] == "fn"]
    bundle_nodes = [n for n in component if nodes[n]["kind"] == "bundle"]
    for n in fn_nodes:
        label = nodes[n]["label"].replace('"', "'")
        lines.append(f'  {abs(hash(n))}["{label}"]')
    for n in bundle_nodes:
        label = nodes[n]["label"].replace('"', "'")
        lines.append(f'  {abs(hash(n))}(({label}))')
    for n in component:
        for nxt in adj.get(n, ()):
            if nxt in component and nodes[n]["kind"] == "fn":
                lines.append(f"  {abs(hash(n))} --> {abs(hash(nxt))}")
    lines.append("  classDef fn fill:#cfe8ff,stroke:#2b6cb0,stroke-width:1px;")
    lines.append("  classDef bundle fill:#ffe9c6,stroke:#c05621,stroke-width:1px;")
    if fn_nodes:
        lines.append(
            "  class "
            + ",".join(str(abs(hash(n))) for n in fn_nodes)
            + " fn;"
        )
    if bundle_nodes:
        lines.append(
            "  class "
            + ",".join(str(abs(hash(n))) for n in bundle_nodes)
            + " bundle;"
        )
    lines.append("```")

    observed = [bundle_map[n] for n in bundle_nodes if n in bundle_map]
    bundle_counts: dict[tuple[str, ...], int] = defaultdict(int)
    for bundle in observed:
        bundle_counts[tuple(sorted(bundle))] += 1
    component_paths: set[Path] = set()
    for n in fn_nodes:
        parts = n.split("::", 2)
        if len(parts) == 3:
            component_paths.add(Path(parts[1]))
    declared = set()
    documented = set()
    for path in component_paths:
        config_path = path.parent / "config.py"
        bundles = config_bundles_by_path.get(config_path)
        if bundles:
            for fields in bundles.values():
                declared.add(tuple(sorted(fields)))
        documented |= documented_bundles_by_path.get(path, set())
    observed_norm = {tuple(sorted(b)) for b in observed}
    observed_only = sorted(observed_norm - declared) if declared else sorted(observed_norm)
    declared_only = sorted(declared - observed_norm)
    documented_only = sorted(observed_norm & documented)
    def _tier(bundle: tuple[str, ...]) -> str:
        count = bundle_counts.get(bundle, 1)
        if declared:
            return "tier-1"
        if count > 1:
            return "tier-2"
        return "tier-3"
    summary_lines = [
        f"Functions: {len(fn_nodes)}",
        f"Observed bundles: {len(observed_norm)}",
    ]
    if not declared:
        summary_lines.append("Declared Config bundles: none found for this component.")
    if observed_only:
        summary_lines.append("Observed-only bundles (not declared in Configs):")
        for bundle in observed_only:
            tier = _tier(bundle)
            documented_flag = "documented" if bundle in documented else "undocumented"
            summary_lines.append(
                f"  - {', '.join(bundle)} ({tier}, {documented_flag})"
            )
    if documented_only:
        summary_lines.append(
            "Documented bundles (dataflow-bundle markers or local dataclass calls):"
        )
        summary_lines.extend(f"  - {', '.join(bundle)}" for bundle in documented_only)
    if declared_only:
        summary_lines.append("Declared Config bundles not observed in this component:")
        summary_lines.extend(f"  - {', '.join(bundle)}" for bundle in declared_only)
    summary = "\n".join(summary_lines)
    return "\n".join(lines), summary


def _emit_report(
    groups_by_path: dict[Path, dict[str, list[set[str]]]],
    max_components: int,
    *,
    type_suggestions: list[str] | None = None,
    type_ambiguities: list[str] | None = None,
    constant_smells: list[str] | None = None,
) -> tuple[str, list[str]]:
    nodes, adj, bundle_map = _component_graph(groups_by_path)
    components = _connected_components(nodes, adj)
    if groups_by_path:
        common = os.path.commonpath([str(p) for p in groups_by_path])
        root = Path(common)
    else:
        root = Path(".")
    config_bundles_by_path = {}
    documented_bundles_by_path = {}
    for path in sorted(root.rglob("config.py")):
        config_bundles_by_path[path] = _iter_config_fields(path)
    protocols = root / "prism_vm_core" / "protocols.py"
    if protocols.exists():
        config_bundles_by_path[protocols] = _iter_config_fields(protocols)
    for path in sorted(root.rglob("*.py")):
        documented = _iter_documented_bundles(path)
        promoted = _iter_dataclass_call_bundles(path)
        documented_bundles_by_path[path] = documented | promoted
    lines = [
        "<!-- dataflow-grammar -->",
        "Dataflow grammar audit (observed forwarding bundles).",
        "",
    ]
    if not components:
        return "\n".join(lines + ["No bundle components detected."]), []
    if len(components) > max_components:
        lines.append(
            f"Showing top {max_components} components of {len(components)}."
        )
    violations: list[str] = []
    for idx, comp in enumerate(components[:max_components], start=1):
        lines.append(f"### Component {idx}")
        mermaid, summary = _render_mermaid_component(
            nodes,
            bundle_map,
            adj,
            comp,
            config_bundles_by_path,
            documented_bundles_by_path,
        )
        lines.append(mermaid)
        lines.append("")
        lines.append("Summary:")
        lines.append("```")
        lines.append(summary)
        lines.append("```")
        lines.append("")
        for line in summary.splitlines():
            if "(tier-3, undocumented)" in line:
                violations.append(line.strip())
            if "(tier-1," in line or "(tier-2," in line:
                if "undocumented" in line:
                    violations.append(line.strip())
    if violations:
        lines.append("Violations:")
        lines.append("```")
        lines.extend(violations)
        lines.append("```")
    if type_suggestions or type_ambiguities:
        lines.append("Type-flow audit:")
        if type_suggestions or type_ambiguities:
            lines.append(_render_type_mermaid(type_suggestions or [], type_ambiguities or []))
        if type_suggestions:
            lines.append("Type tightening candidates:")
            lines.append("```")
            lines.extend(type_suggestions)
            lines.append("```")
        if type_ambiguities:
            lines.append("Type ambiguities (conflicting downstream expectations):")
            lines.append("```")
            lines.extend(type_ambiguities)
            lines.append("```")
    if constant_smells:
        lines.append("Constant-propagation smells (non-test call sites):")
        lines.append("```")
        lines.extend(constant_smells)
        lines.append("```")
    return "\n".join(lines), violations


def _render_type_mermaid(
    suggestions: list[str],
    ambiguities: list[str],
) -> str:
    lines = ["```mermaid", "flowchart LR"]
    node_id = 0
    def _node(label: str) -> str:
        nonlocal node_id
        node_id += 1
        node = f"type_{node_id}"
        safe = label.replace('"', "'")
        lines.append(f'  {node}["{safe}"]')
        return node

    for entry in suggestions:
        # Format: file:func.param can tighten to Type
        if " can tighten to " not in entry:
            continue
        lhs, rhs = entry.split(" can tighten to ", 1)
        src = _node(lhs)
        dst = _node(rhs)
        lines.append(f"  {src} --> {dst}")
    for entry in ambiguities:
        if " downstream types conflict: " not in entry:
            continue
        lhs, rhs = entry.split(" downstream types conflict: ", 1)
        src = _node(lhs)
        # rhs is a repr of list; keep as string nodes per type
        rhs = rhs.strip()
        if rhs.startswith("[") and rhs.endswith("]"):
            rhs = rhs[1:-1]
        type_names = []
        for item in rhs.split(","):
            item = item.strip()
            if not item:
                continue
            item = item.strip("'\"")
            type_names.append(item)
        for type_name in type_names:
            dst = _node(type_name)
            lines.append(f"  {src} -.-> {dst}")
    lines.append("```")
    return "\n".join(lines)


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument("paths", nargs="+")
    parser.add_argument("--no-recursive", action="store_true")
    parser.add_argument("--dot", default=None, help="Write DOT graph to file or '-' for stdout.")
    parser.add_argument("--report", default=None, help="Write Markdown report (mermaid) to file.")
    parser.add_argument("--max-components", type=int, default=10, help="Max components in report.")
    parser.add_argument(
        "--type-audit",
        action="store_true",
        help="Emit type-tightening suggestions based on downstream annotations.",
    )
    parser.add_argument(
        "--type-audit-max",
        type=int,
        default=50,
        help="Max type-tightening entries to print.",
    )
    parser.add_argument(
        "--type-audit-report",
        action="store_true",
        help="Include type-flow audit summary in the markdown report.",
    )
    parser.add_argument(
        "--fail-on-violations",
        action="store_true",
        help="Exit non-zero if undocumented/undeclared bundle violations are detected.",
    )
    args = parser.parse_args()
    paths = _iter_paths(args.paths)
    groups_by_path: dict[Path, dict[str, list[set[str]]]] = {}
    for path in paths:
        groups_by_path[path] = analyze_file(path, recursive=not args.no_recursive)
    if args.dot is not None:
        dot = _emit_dot(groups_by_path)
        if args.dot.strip() == "-":
            print(dot)
        else:
            Path(args.dot).write_text(dot)
        if args.report is None:
            return
    type_suggestions: list[str] = []
    type_ambiguities: list[str] = []
    if args.type_audit or args.type_audit_report:
        type_suggestions, type_ambiguities = analyze_type_flow_repo(paths)
        if args.type_audit:
            if type_suggestions:
                print("Type tightening candidates:")
                for line in type_suggestions[: args.type_audit_max]:
                    print(f"- {line}")
            if type_ambiguities:
                print("Type ambiguities (conflicting downstream expectations):")
                for line in type_ambiguities[: args.type_audit_max]:
                    print(f"- {line}")
            return
        if args.type_audit_report:
            type_suggestions = type_suggestions[: args.type_audit_max]
            type_ambiguities = type_ambiguities[: args.type_audit_max]
    if args.report is not None:
        constant_smells = analyze_constant_flow_repo(paths)
        report, violations = _emit_report(
            groups_by_path,
            args.max_components,
            type_suggestions=type_suggestions if args.type_audit_report else None,
            type_ambiguities=type_ambiguities if args.type_audit_report else None,
            constant_smells=constant_smells,
        )
        Path(args.report).write_text(report)
        if args.fail_on_violations and violations:
            raise SystemExit(
                f"dataflow grammar violations detected: {len(violations)}"
            )
        return
    for path, groups in groups_by_path.items():
        print(f"# {path}")
        for fn, bundles in groups.items():
            if not bundles:
                continue
            print(f"{fn}:")
            for bundle in bundles:
                print(f"  bundle: {sorted(bundle)}")
        print()


if __name__ == "__main__":
    main()



from __future__ import annotations

from pathlib import Path
import re
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.refactor.engine import RefactorEngine
    from gabion.refactor.model import FieldSpec, RefactorRequest

    return RefactorEngine, FieldSpec, RefactorRequest


def _normalize(text: str) -> str:
    return re.sub(r"\s+", "", text)


def test_refactor_engine_emits_protocol_stub(tmp_path: Path) -> None:
    RefactorEngine, FieldSpec, RefactorRequest = _load()
    target = tmp_path / "sample.py"
    target.write_text(
        textwrap.dedent(
            """
            def sink(x, y):
                return x, y
            """
        ).strip()
        + "\n"
    )
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["alpha", "beta"],
        target_path=str(target),
        rationale="Unit test",
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.edits
    replacement = plan.edits[0].replacement
    assert "class BundleProtocol" in replacement
    assert "alpha: object" in replacement
    assert "beta: object" in replacement


def test_refactor_engine_preserves_type_hints(tmp_path: Path) -> None:
    RefactorEngine, FieldSpec, RefactorRequest = _load()
    target = tmp_path / "sample.py"
    target.write_text(
        textwrap.dedent(
            """
            def sink(x, y):
                return x, y
            """
        ).strip()
        + "\n"
    )
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["alpha", "beta"],
        fields=[
            FieldSpec(name="alpha", type_hint="int"),
            FieldSpec(name="beta", type_hint="str"),
        ],
        target_path=str(target),
        rationale="Unit test",
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.edits
    replacement = plan.edits[0].replacement
    assert "alpha: int" in replacement
    assert "beta: str" in replacement


def test_refactor_engine_rewrites_signature_and_preamble(tmp_path: Path) -> None:
    RefactorEngine, FieldSpec, RefactorRequest = _load()
    target = tmp_path / "sample.py"
    target.write_text(
        textwrap.dedent(
            """
            def foo(a, b):
                return a + b
            """
        ).strip()
        + "\n"
    )
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["a", "b"],
        fields=[
            FieldSpec(name="a", type_hint="int"),
            FieldSpec(name="b", type_hint="int"),
        ],
        target_path=str(target),
        target_functions=["foo"],
        rationale="Unit test",
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.edits
    replacement = plan.edits[0].replacement
    assert "def foo(bundle: BundleProtocol)" in replacement
    assert "a = bundle.a" in replacement
    assert "b = bundle.b" in replacement


def test_refactor_engine_rewrites_call_sites(tmp_path: Path) -> None:
    RefactorEngine, FieldSpec, RefactorRequest = _load()
    target = tmp_path / "sample.py"
    target.write_text(
        textwrap.dedent(
            """
            def foo(a, b):
                return a + b

            def caller(x, y):
                return foo(x, y)
            """
        ).strip()
        + "\n"
    )
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["a", "b"],
        fields=[
            FieldSpec(name="a", type_hint="int"),
            FieldSpec(name="b", type_hint="int"),
        ],
        target_path=str(target),
        target_functions=["foo"],
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.edits
    replacement = _normalize(plan.edits[0].replacement)
    assert "returnfoo(BundleProtocol(a=x,b=y))" in replacement


def test_refactor_engine_rewrites_imported_call_sites(tmp_path: Path) -> None:
    RefactorEngine, FieldSpec, RefactorRequest = _load()
    src_root = tmp_path / "src" / "pkg"
    src_root.mkdir(parents=True)
    target = src_root / "mod.py"
    target.write_text(
        textwrap.dedent(
            """
            def foo(a, b):
                return a + b
            """
        ).strip()
        + "\n"
    )
    caller = src_root / "caller.py"
    caller.write_text(
        textwrap.dedent(
            """
            from pkg.mod import foo

            def run(x, y):
                return foo(x, y)
            """
        ).strip()
        + "\n"
    )
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["a", "b"],
        fields=[
            FieldSpec(name="a", type_hint="int"),
            FieldSpec(name="b", type_hint="int"),
        ],
        target_path=str(target),
        target_functions=["foo"],
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.edits
    edits_by_path = {Path(edit.path).name: edit.replacement for edit in plan.edits}
    assert "caller.py" in edits_by_path
    caller_replacement = _normalize(edits_by_path["caller.py"])
    assert "frompkg.modimportBundleProtocol" in caller_replacement
    assert "returnfoo(BundleProtocol(a=x,b=y))" in caller_replacement


def test_refactor_engine_emits_compat_shim(tmp_path: Path) -> None:
    RefactorEngine, FieldSpec, RefactorRequest = _load()
    target = tmp_path / "sample.py"
    target.write_text(
        textwrap.dedent(
            """
            def foo(a, b):
                return a + b
            """
        ).strip()
        + "\n"
    )
    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=["a", "b"],
        fields=[
            FieldSpec(name="a", type_hint="int"),
            FieldSpec(name="b", type_hint="int"),
        ],
        target_path=str(target),
        target_functions=["foo"],
        compatibility_shim=True,
    )
    plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert plan.edits
    replacement = plan.edits[0].replacement
    assert "def _foo_bundle" in replacement
    assert "def foo(*args, **kwargs)" in replacement
    assert "warnings.warn" in replacement
    assert "@overload" in replacement



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _write(tmp_path: Path, rel: str, content: str) -> Path:
    path = tmp_path / rel
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(textwrap.dedent(content).lstrip())
    return path


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_paths, compute_violations, AuditConfig

    return analyze_paths, compute_violations, AuditConfig


def test_config_dataclass_declares_bundle_outside_config_py(tmp_path: Path) -> None:
    analyze_paths, compute_violations, AuditConfig = _load()
    _write(
        tmp_path,
        "settings.py",
        """
        from dataclasses import dataclass

        @dataclass
        class AppConfig:
            a: int
            b: int
        """,
    )
    mod_path = _write(
        tmp_path,
        "mod.py",
        """
        def sink(x=None, y=None):
            return x, y

        def g(a, b):
            sink(x=a, y=a)
            sink(x=b, y=b)

        def h(c, d):
            sink(x=c, y=c)
            sink(x=d, y=d)
        """,
    )
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [tmp_path / "settings.py", mod_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    violations = compute_violations(
        analysis.groups_by_path,
        max_components=10,
    )
    joined = "\n".join(violations)
    assert "a, b" not in joined
    assert "c, d" in joined



#!/usr/bin/env python3
"""Smoke test for the LSP command path."""
from __future__ import annotations

import argparse
import sys
from pathlib import Path

try:
    from gabion.lsp_client import CommandRequest, run_command
except ModuleNotFoundError:
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.lsp_client import CommandRequest, run_command


DATAFLOW_COMMAND = "gabion.dataflowAudit"
SYNTHESIS_COMMAND = "gabion.synthesisPlan"


def main() -> int:
    parser = argparse.ArgumentParser()
    parser.add_argument("--root", default=".")
    args = parser.parse_args()

    payload = {
        "paths": [args.root],
        "fail_on_violations": False,
    }
    result = run_command(
        CommandRequest(DATAFLOW_COMMAND, [payload]), root=Path(args.root)
    )
    if "exit_code" not in result:
        raise SystemExit("Missing exit_code in LSP result")
    synth_payload = {
        "bundles": [{"bundle": ["ctx"], "tier": 2}],
        "min_bundle_size": 1,
        "allow_singletons": True,
        "existing_names": ["CtxBundle"],
    }
    synth_result = run_command(
        CommandRequest(SYNTHESIS_COMMAND, [synth_payload]), root=Path(args.root)
    )
    if "protocols" not in synth_result:
        raise SystemExit("Missing protocols in synthesis result")
    return int(result.get("exit_code", 0))


if __name__ == "__main__":
    raise SystemExit(main())



#!/usr/bin/env python3
"""Docflow audit for governance documents.

Treats frontmatter as the semantic signature and the body as presentation.
Validates required cross-references and commutation symmetry.
"""
from __future__ import annotations

import argparse
from pathlib import Path
import sys
from typing import Dict, List, Tuple

CORE_GOVERNANCE_DOCS = [
    "POLICY_SEED.md",
    "glossary.md",
    "README.md",
    "CONTRIBUTING.md",
    "AGENTS.md",
]

GOVERNANCE_DOCS = CORE_GOVERNANCE_DOCS + [
    "docs/publishing_practices.md",
    "docs/influence_index.md",
    "docs/coverage_semantics.md",
]

REQUIRED_FIELDS = [
    "doc_id",
    "doc_role",
    "doc_scope",
    "doc_authority",
    "doc_requires",
    "doc_reviewed_as_of",
    "doc_change_protocol",
]
LIST_FIELDS = {
    "doc_scope",
    "doc_requires",
    "doc_commutes_with",
    "doc_invariants",
    "doc_erasure",
}
MAP_FIELDS = {
    "doc_reviewed_as_of",
}


def _parse_frontmatter(text: str) -> Tuple[Dict[str, object], str]:
    if not text.startswith("---\n"):
        return {}, text
    lines = text.split("\n")
    if lines[0].strip() != "---":
        return {}, text
    fm_lines: List[str] = []
    idx = 1
    while idx < len(lines):
        line = lines[idx]
        if line.strip() == "---":
            idx += 1
            break
        fm_lines.append(line)
        idx += 1
    body = "\n".join(lines[idx:])
    return _parse_yaml_like(fm_lines), body


def _parse_yaml_like(lines: List[str]) -> Dict[str, object]:
    data: Dict[str, object] = {}
    current_list_key: str | None = None
    current_map_key: str | None = None
    for raw in lines:
        line = raw.rstrip()
        if not line.strip():
            continue
        if line.lstrip().startswith("#"):
            continue
        if current_map_key is not None and line.startswith("  ") and ":" in line:
            key, value = line.strip().split(":", 1)
            key = key.strip()
            value = value.strip()
            if value.startswith("\"") and value.endswith("\""):
                value = value[1:-1]
            if value.isdigit():
                parsed: object = int(value)
            else:
                parsed = value
            mapping = data.get(current_map_key)
            if not isinstance(mapping, dict):
                mapping = {}
            mapping[key] = parsed
            data[current_map_key] = mapping
            continue
        if line.lstrip().startswith("- "):
            if current_list_key is None:
                continue
            item = line.strip()[2:].strip()
            if isinstance(data.get(current_list_key), list):
                data[current_list_key].append(item)
            continue
        if ":" in line:
            key, value = line.split(":", 1)
            key = key.strip()
            value = value.strip()
            if value == "":
                if key in MAP_FIELDS:
                    data[key] = {}
                    current_map_key = key
                    current_list_key = None
                else:
                    data[key] = []
                    current_list_key = key
                    current_map_key = None
                continue
            if value.startswith("\"") and value.endswith("\""):
                value = value[1:-1]
            if value.isdigit():
                data[key] = int(value)
            else:
                data[key] = value
            current_list_key = None
            current_map_key = None
    return data


def _lower_name(path: Path) -> str:
    return path.stem.lower()


def _audit(root: Path) -> Tuple[List[str], List[str]]:
    violations: List[str] = []
    warnings: List[str] = []

    docs: Dict[str, Dict[str, object]] = {}
    doc_ids: Dict[str, str] = {}

    for rel in GOVERNANCE_DOCS:
        path = root / rel
        if not path.exists():
            violations.append(f"missing governance doc: {rel}")
            continue
        text = path.read_text(encoding="utf-8")
        fm, body = _parse_frontmatter(text)
        docs[rel] = {"frontmatter": fm, "body": body}
        doc_id = fm.get("doc_id")
        if isinstance(doc_id, str):
            if doc_id in doc_ids:
                violations.append(
                    f"duplicate doc_id '{doc_id}' in {rel} and {doc_ids[doc_id]}"
                )
            else:
                doc_ids[doc_id] = rel

    governance_set = set(GOVERNANCE_DOCS)
    core_set = set(CORE_GOVERNANCE_DOCS)

    revisions: Dict[str, int] = {}
    for rel, payload in docs.items():
        fm = payload["frontmatter"]
        if isinstance(fm.get("doc_revision"), int):
            revisions[rel] = fm["doc_revision"]

    for rel, payload in docs.items():
        fm = payload["frontmatter"]
        body = payload["body"]
        # Required fields
        for field in REQUIRED_FIELDS:
            if field not in fm:
                violations.append(f"{rel}: missing frontmatter field '{field}'")
        # Required list for doc_scope/doc_requires
        for field in ("doc_scope", "doc_requires"):
            if field in fm and not isinstance(fm[field], list):
                violations.append(f"{rel}: frontmatter field '{field}' must be a list")
        if "doc_reviewed_as_of" in fm and not isinstance(fm["doc_reviewed_as_of"], dict):
            violations.append(f"{rel}: frontmatter field 'doc_reviewed_as_of' must be a map")
        # Normative docs must require the governance bundle.
        if fm.get("doc_authority") == "normative":
            requires = set(fm.get("doc_requires", []))
            expected = core_set - {rel}
            missing = sorted(expected - requires)
            if missing:
                violations.append(
                    f"{rel}: missing required governance references: {', '.join(missing)}"
                )
        # Commutation symmetry
        commutes = fm.get("doc_commutes_with")
        if isinstance(commutes, list):
            for other in commutes:
                other_fm = docs.get(other, {}).get("frontmatter", {})
                other_commutes = other_fm.get("doc_commutes_with", [])
                if other not in docs:
                    violations.append(f"{rel}: doc_commutes_with target missing: {other}")
                    continue
                if not isinstance(other_commutes, list) or rel not in other_commutes:
                    violations.append(
                        f"{rel}: commutation with {other} not reciprocated"
                    )
        # Body references vs frontmatter requirements
        requires = fm.get("doc_requires", [])
        if isinstance(requires, list):
            body_lower = body.lower()
            for req in requires:
                if req in body:
                    continue
                req_name = _lower_name(Path(req))
                if req_name and req_name in body_lower:
                    warnings.append(
                        f"{rel}: implicit reference to {req} (Tier-2); prefer explicit path"
                    )
                else:
                    violations.append(f"{rel}: missing explicit reference to {req}")
        # Convergence check (re-reviewed as of)
        reviewed = fm.get("doc_reviewed_as_of")
        if isinstance(requires, list) and requires:
            if not isinstance(reviewed, dict):
                violations.append(f"{rel}: doc_reviewed_as_of missing or invalid")
            else:
                for req in requires:
                    expected = revisions.get(req)
                    if expected is None:
                        violations.append(
                            f"{rel}: doc_reviewed_as_of cannot resolve {req}"
                        )
                        continue
                    seen = reviewed.get(req)
                    if not isinstance(seen, int):
                        violations.append(
                            f"{rel}: doc_reviewed_as_of[{req}] must be an integer"
                        )
                    elif seen != expected:
                        violations.append(
                            f"{rel}: doc_reviewed_as_of[{req}]={seen} does not match {expected}"
                        )

    warnings.extend(_tooling_warnings(root, docs))
    warnings.extend(_influence_warnings(root))

    return violations, warnings


def _tooling_warnings(root: Path, docs: Dict[str, Dict[str, object]]) -> List[str]:
    warnings: List[str] = []
    makefile = root / "Makefile"
    if makefile.exists():
        for rel in ("README.md", "CONTRIBUTING.md"):
            body = docs.get(rel, {}).get("body", "")
            if "make " not in body and "Make targets" not in body:
                warnings.append(
                    f"{rel}: Makefile present but make targets are not documented"
                )
    checks_script = root / "scripts" / "checks.sh"
    if checks_script.exists():
        body = docs.get("CONTRIBUTING.md", {}).get("body", "")
        if "scripts/checks.sh" not in body:
            warnings.append(
                "CONTRIBUTING.md: scripts/checks.sh present but not documented"
            )
    return warnings


def _influence_warnings(root: Path) -> List[str]:
    warnings: List[str] = []
    inbox = root / "in"
    index_path = root / "docs" / "influence_index.md"
    if not inbox.exists():
        return warnings
    if not index_path.exists():
        warnings.append("docs/influence_index.md: missing influence index for in/")
        return warnings
    index_text = index_path.read_text(encoding="utf-8")
    for path in sorted(inbox.glob("in-*.md")):
        rel = path.as_posix()
        if rel not in index_text:
            warnings.append(f"docs/influence_index.md: missing {rel}")
    return warnings


def main() -> None:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--root",
        default=".",
        help="Repository root (default: current directory)",
    )
    parser.add_argument(
        "--fail-on-violations",
        action="store_true",
        help="Exit non-zero if violations are detected",
    )
    args = parser.parse_args()
    root = Path(args.root)

    violations, warnings = _audit(root)

    print("Docflow audit summary")
    if warnings:
        print("Warnings:")
        for w in warnings:
            print(f"- {w}")
    if violations:
        print("Violations:")
        for v in violations:
            print(f"- {v}")
    if not warnings and not violations:
        print("No issues detected.")

    if violations and args.fail_on_violations:
        raise SystemExit(1)


if __name__ == "__main__":
    main()



from __future__ import annotations

from typing import Dict, List, Optional, Tuple

from pydantic import BaseModel


class BundleDTO(BaseModel):
    fields: List[str]
    locations: List[str]
    suggested_name: str
    tier: int


class AnalysisRequest(BaseModel):
    root_path: str


class AnalysisResponse(BaseModel):
    bundles: List[BundleDTO]
    stats: Dict[str, int]


class SynthesisBundleDTO(BaseModel):
    bundle: List[str]
    tier: int


class SynthesisRequest(BaseModel):
    bundles: List[SynthesisBundleDTO]
    field_types: Dict[str, str] = {}
    existing_names: List[str] = []
    frequency: Dict[str, int] = {}
    fallback_prefix: str = "Bundle"
    max_tier: int = 2
    min_bundle_size: int = 2
    allow_singletons: bool = False
    merge_overlap_threshold: float = 0.75


class SynthesisFieldDTO(BaseModel):
    name: str
    type_hint: Optional[str] = None
    source_params: List[str] = []


class SynthesisProtocolDTO(BaseModel):
    name: str
    fields: List[SynthesisFieldDTO]
    bundle: List[str]
    tier: int
    rationale: Optional[str] = None


class SynthesisResponse(BaseModel):
    protocols: List[SynthesisProtocolDTO]
    warnings: List[str] = []
    errors: List[str] = []


class RefactorFieldDTO(BaseModel):
    name: str
    type_hint: Optional[str] = None


class RefactorRequest(BaseModel):
    protocol_name: str
    bundle: List[str]
    fields: List[RefactorFieldDTO] = []
    target_path: str
    target_functions: List[str] = []
    compatibility_shim: bool = False
    rationale: Optional[str] = None


class TextEditDTO(BaseModel):
    path: str
    start: Tuple[int, int]
    end: Tuple[int, int]
    replacement: str


class RefactorResponse(BaseModel):
    edits: List[TextEditDTO] = []
    warnings: List[str] = []
    errors: List[str] = []



from __future__ import annotations

import json
import subprocess
import sys
from pathlib import Path
from dataclasses import dataclass
from typing import Any, Callable


class LspClientError(RuntimeError):
    pass


@dataclass(frozen=True)
class CommandRequest:
    command: str
    arguments: list[dict] | None = None


def _read_rpc(stream) -> dict:
    header = b""
    while b"\r\n\r\n" not in header:
        chunk = stream.read(1)
        if not chunk:
            raise LspClientError("LSP stream closed")
        header += chunk
    head, _, rest = header.partition(b"\r\n\r\n")
    length = 0
    for line in head.split(b"\r\n"):
        if line.lower().startswith(b"content-length:"):
            length = int(line.split(b":", 1)[1].strip())
            break
    if length <= 0:
        raise LspClientError("Invalid LSP Content-Length")
    body = rest
    if len(body) < length:
        body += stream.read(length - len(body))
    return json.loads(body.decode("utf-8"))


def _write_rpc(stream, message: dict) -> None:
    payload = json.dumps(message).encode("utf-8")
    header = f"Content-Length: {len(payload)}\r\n\r\n".encode("utf-8")
    stream.write(header + payload)
    stream.flush()


def _read_response(stream, request_id: int) -> dict:
    while True:
        message = _read_rpc(stream)
        if message.get("id") == request_id:
            return message


def run_command(
    request: CommandRequest,
    *,
    root: Path | None = None,
    timeout: float = 5.0,
    process_factory: Callable[..., subprocess.Popen] = subprocess.Popen,
) -> dict[str, Any]:
    proc = process_factory(
        [sys.executable, "-m", "gabion.server"],
        stdin=subprocess.PIPE,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
    )
    assert proc.stdin is not None
    assert proc.stdout is not None

    root_uri = (root or Path.cwd()).resolve().as_uri()
    initialize_id = 1
    _write_rpc(
        proc.stdin,
        {
            "jsonrpc": "2.0",
            "id": initialize_id,
            "method": "initialize",
            "params": {"rootUri": root_uri, "capabilities": {}},
        },
    )
    _read_response(proc.stdout, initialize_id)
    _write_rpc(proc.stdin, {"jsonrpc": "2.0", "method": "initialized", "params": {}})

    cmd_id = 2
    _write_rpc(
        proc.stdin,
        {
            "jsonrpc": "2.0",
            "id": cmd_id,
            "method": "workspace/executeCommand",
            "params": {"command": request.command, "arguments": request.arguments or []},
        },
    )
    response = _read_response(proc.stdout, cmd_id)

    shutdown_id = 3
    _write_rpc(proc.stdin, {"jsonrpc": "2.0", "id": shutdown_id, "method": "shutdown"})
    _read_response(proc.stdout, shutdown_id)
    _write_rpc(proc.stdin, {"jsonrpc": "2.0", "method": "exit"})
    out, err = proc.communicate(timeout=timeout)
    if response.get("error"):
        raise LspClientError(f"LSP error: {response['error']}")
    if proc.returncode not in (0, None):
        detail = err.decode("utf-8", errors="replace").strip()
        raise LspClientError(f"LSP server failed (exit {proc.returncode}): {detail}")
    if err:
        detail = err.decode("utf-8", errors="replace").strip()
        if detail:
            raise LspClientError(f"LSP server error output: {detail}")
    return response.get("result", {})



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import render_protocol_stubs

    return render_protocol_stubs


def test_render_protocol_stubs_emits_dataclass() -> None:
    render_protocol_stubs = _load()
    plan = {
        "protocols": [
            {
                "name": "ExampleBundle",
                "fields": [
                    {"name": "ctx", "type_hint": "Context"},
                    {"name": "config", "type_hint": None},
                ],
            }
        ]
    }
    stub = render_protocol_stubs(plan)
    assert "from typing import Any" in stub
    assert "from dataclasses import dataclass" in stub
    assert "class TODO_Name_Me" in stub
    assert "ctx: Context" in stub
    assert "config: Any" in stub


def test_render_protocol_stubs_emits_protocol() -> None:
    render_protocol_stubs = _load()
    plan = {
        "protocols": [
            {
                "name": "ExampleBundle",
                "fields": [
                    {"name": "ctx", "type_hint": "Context"},
                    {"name": "config", "type_hint": None},
                ],
            }
        ]
    }
    stub = render_protocol_stubs(plan, kind="protocol")
    assert "from dataclasses import dataclass" not in stub
    assert "from typing import Any, Protocol" in stub
    assert "class TODO_Name_Me(Protocol)" in stub
    assert "ctx: Context" in stub
    assert "config: Any" in stub



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from typing import List, Optional, Any, Callable
import argparse
import json
import subprocess
import sys

import typer

DATAFLOW_COMMAND = "gabion.dataflowAudit"
SYNTHESIS_COMMAND = "gabion.synthesisPlan"
REFACTOR_COMMAND = "gabion.refactorProtocol"
from gabion.lsp_client import CommandRequest, run_command
app = typer.Typer(add_completion=False)


@dataclass(frozen=True)
class DataflowAuditRequest:
    ctx: typer.Context
    args: List[str] | None = None
    runner: Callable[..., dict[str, Any]] | None = None


def _find_repo_root() -> Path:
    return Path(__file__).resolve().parents[2]


def _split_csv_entries(entries: Optional[List[str]]) -> list[str] | None:
    if entries is None:
        return None
    merged: list[str] = []
    for entry in entries:
        merged.extend([part.strip() for part in entry.split(",") if part.strip()])
    return merged or None


def _split_csv(value: Optional[str]) -> list[str] | None:
    if value is None:
        return None
    items = [part.strip() for part in value.split(",") if part.strip()]
    return items or None


def build_check_payload(
    *,
    paths: Optional[List[Path]],
    report: Optional[Path],
    fail_on_violations: bool,
    root: Path | None,
    config: Optional[Path],
    baseline: Optional[Path],
    baseline_write: bool,
    exclude: Optional[List[str]],
    ignore_params: Optional[str],
    transparent_decorators: Optional[str],
    allow_external: Optional[bool],
    strictness: Optional[str],
    fail_on_type_ambiguities: bool,
) -> dict[str, Any]:
    # dataflow-bundle: ignore_params, transparent_decorators
    if not paths:
        paths = [Path(".")]
    if strictness is not None and strictness not in {"high", "low"}:
        raise typer.BadParameter("strictness must be 'high' or 'low'")
    exclude_dirs = _split_csv_entries(exclude)
    ignore_list = _split_csv(ignore_params)
    transparent_list = _split_csv(transparent_decorators)
    baseline_write_value: bool | None = baseline_write if baseline is not None else None
    root = root or Path(".")
    payload = {
        "paths": [str(p) for p in paths],
        "report": str(report) if report is not None else None,
        "fail_on_violations": fail_on_violations,
        "fail_on_type_ambiguities": fail_on_type_ambiguities,
        "root": str(root),
        "config": str(config) if config is not None else None,
        "baseline": str(baseline) if baseline is not None else None,
        "baseline_write": baseline_write_value,
        "exclude": exclude_dirs,
        "ignore_params": ignore_list,
        "transparent_decorators": transparent_list,
        "allow_external": allow_external,
        "strictness": strictness,
        "type_audit": True if fail_on_type_ambiguities else None,
    }
    return payload


def parse_dataflow_args(argv: list[str]) -> argparse.Namespace:
    parser = dataflow_cli_parser()
    return parser.parse_args(argv)


def build_dataflow_payload(opts: argparse.Namespace) -> dict[str, Any]:
    exclude_dirs = _split_csv_entries(opts.exclude)
    ignore_list = _split_csv(opts.ignore_params)
    transparent_list = _split_csv(opts.transparent_decorators)
    payload: dict[str, Any] = {
        "paths": [str(p) for p in opts.paths],
        "root": str(opts.root),
        "config": str(opts.config) if opts.config is not None else None,
        "report": str(opts.report) if opts.report else None,
        "dot": opts.dot,
        "fail_on_violations": opts.fail_on_violations,
        "fail_on_type_ambiguities": opts.fail_on_type_ambiguities,
        "baseline": str(opts.baseline) if opts.baseline else None,
        "baseline_write": opts.baseline_write if opts.baseline else None,
        "no_recursive": opts.no_recursive,
        "max_components": opts.max_components,
        "type_audit": opts.type_audit,
        "type_audit_report": opts.type_audit_report,
        "type_audit_max": opts.type_audit_max,
        "exclude": exclude_dirs,
        "ignore_params": ignore_list,
        "transparent_decorators": transparent_list,
        "allow_external": opts.allow_external,
        "strictness": opts.strictness,
        "synthesis_plan": str(opts.synthesis_plan) if opts.synthesis_plan else None,
        "synthesis_report": opts.synthesis_report,
        "synthesis_max_tier": opts.synthesis_max_tier,
        "synthesis_min_bundle_size": opts.synthesis_min_bundle_size,
        "synthesis_allow_singletons": opts.synthesis_allow_singletons,
        "synthesis_protocols": str(opts.synthesis_protocols)
        if opts.synthesis_protocols
        else None,
        "synthesis_protocols_kind": opts.synthesis_protocols_kind,
        "refactor_plan": opts.refactor_plan,
        "refactor_plan_json": str(opts.refactor_plan_json)
        if opts.refactor_plan_json
        else None,
        "synthesis_merge_overlap": opts.synthesis_merge_overlap,
    }
    return payload


def build_refactor_payload(
    *,
    input_payload: Optional[dict[str, Any]] = None,
    protocol_name: Optional[str],
    bundle: Optional[List[str]],
    field: Optional[List[str]],
    target_path: Optional[Path],
    target_functions: Optional[List[str]],
    compatibility_shim: bool,
    rationale: Optional[str],
) -> dict[str, Any]:
    if input_payload is not None:
        return input_payload
    if protocol_name is None or target_path is None:
        raise typer.BadParameter(
            "Provide --protocol-name and --target-path or use --input."
        )
    field_specs: list[dict[str, str | None]] = []
    for spec in field or []:
        name, _, hint = spec.partition(":")
        name = name.strip()
        if not name:
            continue
        type_hint = hint.strip() or None
        field_specs.append({"name": name, "type_hint": type_hint})
    if not bundle and field_specs:
        bundle = [spec["name"] for spec in field_specs]
    return {
        "protocol_name": protocol_name,
        "bundle": bundle or [],
        "fields": field_specs,
        "target_path": str(target_path),
        "target_functions": target_functions or [],
        "compatibility_shim": compatibility_shim,
        "rationale": rationale,
    }


def dispatch_command(
    *,
    command: str,
    payload: dict[str, Any],
    root: Path | None = None,
    runner: Callable[..., dict[str, Any]] = run_command,
) -> dict[str, Any]:
    request = CommandRequest(command, [payload])
    return runner(request, root=root)


def run_check(
    *,
    paths: Optional[List[Path]],
    report: Optional[Path],
    fail_on_violations: bool,
    root: Path,
    config: Optional[Path],
    baseline: Optional[Path],
    baseline_write: bool,
    exclude: Optional[List[str]],
    ignore_params: Optional[str],
    transparent_decorators: Optional[str],
    allow_external: Optional[bool],
    strictness: Optional[str],
    fail_on_type_ambiguities: bool,
    runner: Callable[..., dict[str, Any]] = run_command,
) -> dict[str, Any]:
    # dataflow-bundle: ignore_params, transparent_decorators
    payload = build_check_payload(
        paths=paths,
        report=report,
        fail_on_violations=fail_on_violations,
        root=root,
        config=config,
        baseline=baseline,
        baseline_write=baseline_write if baseline is not None else False,
        exclude=exclude,
        ignore_params=ignore_params,
        transparent_decorators=transparent_decorators,
        allow_external=allow_external,
        strictness=strictness,
        fail_on_type_ambiguities=fail_on_type_ambiguities,
    )
    return dispatch_command(command=DATAFLOW_COMMAND, payload=payload, root=root, runner=runner)


@app.command()
def check(
    paths: List[Path] = typer.Argument(None),
    report: Optional[Path] = typer.Option(None, "--report"),
    fail_on_violations: bool = typer.Option(True, "--fail-on-violations/--no-fail-on-violations"),
    root: Path = typer.Option(Path("."), "--root"),
    config: Optional[Path] = typer.Option(None, "--config"),
    baseline: Optional[Path] = typer.Option(
        None, "--baseline", help="Baseline file of allowed violations."
    ),
    baseline_write: bool = typer.Option(
        False, "--baseline-write", help="Write current violations to baseline."
    ),
    exclude: Optional[List[str]] = typer.Option(None, "--exclude"),
    ignore_params: Optional[str] = typer.Option(None, "--ignore-params"),
    transparent_decorators: Optional[str] = typer.Option(
        None, "--transparent-decorators"
    ),
    allow_external: Optional[bool] = typer.Option(
        None, "--allow-external/--no-allow-external"
    ),
    strictness: Optional[str] = typer.Option(None, "--strictness"),
    fail_on_type_ambiguities: bool = typer.Option(
        True, "--fail-on-type-ambiguities/--no-fail-on-type-ambiguities"
    ),
) -> None:
    # dataflow-bundle: ignore_params, transparent_decorators
    """Run the dataflow grammar audit with strict defaults."""
    result = run_check(
        paths=paths,
        report=report,
        fail_on_violations=fail_on_violations,
        root=root,
        config=config,
        baseline=baseline,
        baseline_write=baseline_write,
        exclude=exclude,
        ignore_params=ignore_params,
        transparent_decorators=transparent_decorators,
        allow_external=allow_external,
        strictness=strictness,
        fail_on_type_ambiguities=fail_on_type_ambiguities,
    )
    raise typer.Exit(code=int(result.get("exit_code", 0)))


def _dataflow_audit(
    request: "DataflowAuditRequest",
) -> None:
    """Run the dataflow grammar audit with explicit options."""
    argv = list(request.args or []) + list(request.ctx.args)
    if not argv:
        argv = []
    opts = parse_dataflow_args(argv)
    payload = build_dataflow_payload(opts)
    runner = request.runner or run_command
    result = dispatch_command(
        command=DATAFLOW_COMMAND,
        payload=payload,
        root=Path(opts.root),
        runner=runner,
    )
    if opts.type_audit:
        suggestions = result.get("type_suggestions", [])
        ambiguities = result.get("type_ambiguities", [])
        if suggestions:
            typer.echo("Type tightening candidates:")
            for line in suggestions[: opts.type_audit_max]:
                typer.echo(f"- {line}")
        if ambiguities:
            typer.echo("Type ambiguities (conflicting downstream expectations):")
            for line in ambiguities[: opts.type_audit_max]:
                typer.echo(f"- {line}")
    if opts.dot == "-" and "dot" in result:
        typer.echo(result["dot"])
    if opts.synthesis_plan == "-" and "synthesis_plan" in result:
        typer.echo(json.dumps(result["synthesis_plan"], indent=2, sort_keys=True))
    if opts.synthesis_protocols == "-" and "synthesis_protocols" in result:
        typer.echo(result["synthesis_protocols"])
    if opts.refactor_plan_json == "-" and "refactor_plan" in result:
        typer.echo(json.dumps(result["refactor_plan"], indent=2, sort_keys=True))
    raise typer.Exit(code=int(result.get("exit_code", 0)))


@app.command(
    "dataflow-audit",
    context_settings={"allow_extra_args": True, "ignore_unknown_options": True},
)
def dataflow_audit(
    ctx: typer.Context,
    args: List[str] = typer.Argument(None),
) -> None:
    request = DataflowAuditRequest(ctx=ctx, args=args)
    _dataflow_audit(request)


def dataflow_cli_parser() -> argparse.ArgumentParser:
    parser = argparse.ArgumentParser()
    parser.add_argument("paths", nargs="+")
    parser.add_argument("--root", default=".")
    parser.add_argument("--config", default=None)
    parser.add_argument("--baseline", default=None, help="Baseline file for violations.")
    parser.add_argument(
        "--baseline-write",
        action="store_true",
        help="Write current violations to baseline file.",
    )
    parser.add_argument("--exclude", action="append", default=None)
    parser.add_argument("--ignore-params", default=None)
    parser.add_argument(
        "--transparent-decorators",
        default=None,
        help="Comma-separated decorator names treated as transparent.",
    )
    parser.add_argument(
        "--allow-external",
        action=argparse.BooleanOptionalAction,
        default=None,
    )
    parser.add_argument("--strictness", choices=["high", "low"], default=None)
    parser.add_argument("--no-recursive", action="store_true")
    parser.add_argument("--dot", default=None, help="Write DOT graph to file or '-' for stdout.")
    parser.add_argument("--report", default=None, help="Write Markdown report (mermaid) to file.")
    parser.add_argument("--max-components", type=int, default=10, help="Max components in report.")
    parser.add_argument(
        "--type-audit",
        action="store_true",
        help="Emit type-tightening suggestions based on downstream annotations.",
    )
    parser.add_argument(
        "--type-audit-max",
        type=int,
        default=50,
        help="Max type-tightening entries to print.",
    )
    parser.add_argument(
        "--type-audit-report",
        action="store_true",
        help="Include type-flow audit summary in the markdown report.",
    )
    parser.add_argument(
        "--fail-on-type-ambiguities",
        action="store_true",
        help="Exit non-zero if type ambiguities are detected.",
    )
    parser.add_argument(
        "--fail-on-violations",
        action="store_true",
        help="Exit non-zero if undocumented/undeclared bundle violations are detected.",
    )
    parser.add_argument(
        "--synthesis-plan",
        default=None,
        help="Write synthesis plan JSON to file or '-' for stdout.",
    )
    parser.add_argument(
        "--synthesis-report",
        action="store_true",
        help="Include synthesis plan summary in the markdown report.",
    )
    parser.add_argument(
        "--synthesis-protocols",
        default=None,
        help="Write protocol/dataclass stubs to file or '-' for stdout.",
    )
    parser.add_argument(
        "--synthesis-protocols-kind",
        choices=["dataclass", "protocol"],
        default="dataclass",
        help="Emit dataclass or typing.Protocol stubs (default: dataclass).",
    )
    parser.add_argument(
        "--synthesis-max-tier",
        type=int,
        default=2,
        help="Max tier to include in synthesis plan.",
    )
    parser.add_argument(
        "--synthesis-min-bundle-size",
        type=int,
        default=2,
        help="Min bundle size to include in synthesis plan.",
    )
    parser.add_argument(
        "--synthesis-allow-singletons",
        action="store_true",
        help="Allow single-field bundles in synthesis plan.",
    )
    parser.add_argument(
        "--synthesis-merge-overlap",
        type=float,
        default=None,
        help="Jaccard overlap threshold for merging bundles (0.0-1.0).",
    )
    parser.add_argument(
        "--refactor-plan",
        action="store_true",
        help="Include refactoring plan summary in the markdown report.",
    )
    parser.add_argument(
        "--refactor-plan-json",
        default=None,
        help="Write refactoring plan JSON to file or '-' for stdout.",
    )
    return parser


def _run_docflow_audit(
    *,
    root: Path,
    fail_on_violations: bool,
    script: Path | None = None,
) -> int:
    repo_root = _find_repo_root()
    script_path = script or (repo_root / "scripts" / "docflow_audit.py")
    if not script_path.exists():
        typer.secho(
            "docflow audit script not found; repository layout required",
            err=True,
            fg=typer.colors.RED,
        )
        return 2
    args = ["--root", str(root)]
    if fail_on_violations:
        args.append("--fail-on-violations")
    result = subprocess.run([sys.executable, str(script_path), *args], check=False)
    return result.returncode


@app.command("docflow-audit")
def docflow_audit(
    root: Path = typer.Option(Path("."), "--root"),
    fail_on_violations: bool = typer.Option(
        False, "--fail-on-violations/--no-fail-on-violations"
    ),
) -> None:
    """Run the docflow audit (governance docs only)."""
    exit_code = _run_docflow_audit(root=root, fail_on_violations=fail_on_violations)
    raise typer.Exit(code=exit_code)


def _run_synth(
    *,
    paths: List[Path] | None,
    root: Path,
    out_dir: Path,
    no_timestamp: bool,
    config: Optional[Path],
    exclude: Optional[List[str]],
    ignore_params: Optional[str],
    transparent_decorators: Optional[str],
    allow_external: Optional[bool],
    strictness: Optional[str],
    no_recursive: bool,
    max_components: int,
    type_audit_report: bool,
    type_audit_max: int,
    synthesis_max_tier: int,
    synthesis_min_bundle_size: int,
    synthesis_allow_singletons: bool,
    synthesis_protocols_kind: str,
    refactor_plan: bool,
    fail_on_violations: bool,
    runner: Callable[..., dict[str, Any]] = run_command,
) -> tuple[dict[str, Any], dict[str, Path], Path | None]:
    if not paths:
        paths = [Path(".")]
    exclude_dirs: list[str] | None = None
    if exclude is not None:
        exclude_dirs = []
        for entry in exclude:
            exclude_dirs.extend([part.strip() for part in entry.split(",") if part.strip()])
    ignore_list: list[str] | None = None
    if ignore_params is not None:
        ignore_list = [p.strip() for p in ignore_params.split(",") if p.strip()]
    transparent_list: list[str] | None = None
    if transparent_decorators is not None:
        transparent_list = [
            p.strip() for p in transparent_decorators.split(",") if p.strip()
        ]
    if strictness is not None and strictness not in {"high", "low"}:
        raise typer.BadParameter("strictness must be 'high' or 'low'")
    if synthesis_protocols_kind not in {"dataclass", "protocol"}:
        raise typer.BadParameter(
            "synthesis-protocols-kind must be 'dataclass' or 'protocol'"
        )

    output_root = out_dir
    timestamp = None
    if not no_timestamp:
        timestamp = datetime.now(tz=timezone.utc).strftime("%Y%m%d_%H%M%S")
        output_root = out_dir / timestamp
        out_dir.mkdir(parents=True, exist_ok=True)
        (out_dir / "LATEST.txt").write_text(timestamp)
    output_root.mkdir(parents=True, exist_ok=True)

    report_path = output_root / "dataflow_report.md"
    dot_path = output_root / "dataflow_graph.dot"
    plan_path = output_root / "synthesis_plan.json"
    protocol_path = output_root / "protocol_stubs.py"
    refactor_plan_path = output_root / "refactor_plan.json"

    payload: dict[str, Any] = {
        "paths": [str(p) for p in paths],
        "root": str(root),
        "config": str(config) if config is not None else None,
        "report": str(report_path),
        "dot": str(dot_path),
        "fail_on_violations": fail_on_violations,
        "no_recursive": no_recursive,
        "max_components": max_components,
        "type_audit_report": type_audit_report,
        "type_audit_max": type_audit_max,
        "exclude": exclude_dirs,
        "ignore_params": ignore_list,
        "transparent_decorators": transparent_list,
        "allow_external": allow_external,
        "strictness": strictness,
        "synthesis_plan": str(plan_path),
        "synthesis_report": True,
        "synthesis_protocols": str(protocol_path),
        "synthesis_protocols_kind": synthesis_protocols_kind,
        "synthesis_max_tier": synthesis_max_tier,
        "synthesis_min_bundle_size": synthesis_min_bundle_size,
        "synthesis_allow_singletons": synthesis_allow_singletons,
        "refactor_plan": refactor_plan,
        "refactor_plan_json": str(refactor_plan_path) if refactor_plan else None,
    }
    result = dispatch_command(
        command=DATAFLOW_COMMAND,
        payload=payload,
        root=root,
        runner=runner,
    )
    paths_out = {
        "report": report_path,
        "dot": dot_path,
        "plan": plan_path,
        "protocol": protocol_path,
        "refactor": refactor_plan_path,
        "output_root": output_root,
    }
    return result, paths_out, timestamp


@app.command("synth")
def synth(
    paths: List[Path] = typer.Argument(None),
    root: Path = typer.Option(Path("."), "--root"),
    out_dir: Path = typer.Option(Path("artifacts/synthesis"), "--out-dir"),
    no_timestamp: bool = typer.Option(False, "--no-timestamp"),
    config: Optional[Path] = typer.Option(None, "--config"),
    exclude: Optional[List[str]] = typer.Option(None, "--exclude"),
    ignore_params: Optional[str] = typer.Option(None, "--ignore-params"),
    transparent_decorators: Optional[str] = typer.Option(
        None, "--transparent-decorators"
    ),
    allow_external: Optional[bool] = typer.Option(
        None, "--allow-external/--no-allow-external"
    ),
    strictness: Optional[str] = typer.Option(None, "--strictness"),
    no_recursive: bool = typer.Option(False, "--no-recursive"),
    max_components: int = typer.Option(10, "--max-components"),
    type_audit_report: bool = typer.Option(
        True, "--type-audit-report/--no-type-audit-report"
    ),
    type_audit_max: int = typer.Option(50, "--type-audit-max"),
    synthesis_max_tier: int = typer.Option(2, "--synthesis-max-tier"),
    synthesis_min_bundle_size: int = typer.Option(2, "--synthesis-min-bundle-size"),
    synthesis_allow_singletons: bool = typer.Option(
        False, "--synthesis-allow-singletons"
    ),
    synthesis_protocols_kind: str = typer.Option(
        "dataclass", "--synthesis-protocols-kind"
    ),
    refactor_plan: bool = typer.Option(True, "--refactor-plan/--no-refactor-plan"),
    fail_on_violations: bool = typer.Option(
        False, "--fail-on-violations/--no-fail-on-violations"
    ),
) -> None:
    """Run the dataflow audit and emit synthesis outputs (prototype)."""
    result, paths_out, timestamp = _run_synth(
        paths=paths,
        root=root,
        out_dir=out_dir,
        no_timestamp=no_timestamp,
        config=config,
        exclude=exclude,
        ignore_params=ignore_params,
        transparent_decorators=transparent_decorators,
        allow_external=allow_external,
        strictness=strictness,
        no_recursive=no_recursive,
        max_components=max_components,
        type_audit_report=type_audit_report,
        type_audit_max=type_audit_max,
        synthesis_max_tier=synthesis_max_tier,
        synthesis_min_bundle_size=synthesis_min_bundle_size,
        synthesis_allow_singletons=synthesis_allow_singletons,
        synthesis_protocols_kind=synthesis_protocols_kind,
        refactor_plan=refactor_plan,
        fail_on_violations=fail_on_violations,
    )
    if timestamp:
        typer.echo(f"Snapshot: {paths_out['output_root']}")
    typer.echo(f"- {paths_out['report']}")
    typer.echo(f"- {paths_out['dot']}")
    typer.echo(f"- {paths_out['plan']}")
    typer.echo(f"- {paths_out['protocol']}")
    if refactor_plan:
        typer.echo(f"- {paths_out['refactor']}")
    raise typer.Exit(code=int(result.get("exit_code", 0)))


@app.command("synthesis-plan")
def synthesis_plan(
    input_path: Optional[Path] = typer.Option(
        None, "--input", help="JSON payload describing bundles and synthesis settings."
    ),
    output_path: Optional[Path] = typer.Option(
        None, "--output", help="Write synthesis plan JSON to this path."
    ),
) -> None:
    """Generate a synthesis plan from a JSON payload (prototype)."""
    _run_synthesis_plan(input_path=input_path, output_path=output_path)


def _run_synthesis_plan(
    *,
    input_path: Optional[Path],
    output_path: Optional[Path],
    runner: Callable[..., dict[str, Any]] = run_command,
) -> None:
    """Generate a synthesis plan from a JSON payload (prototype)."""
    payload: dict[str, Any] = {}
    if input_path is not None:
        try:
            payload = json.loads(input_path.read_text())
        except json.JSONDecodeError as exc:
            raise typer.BadParameter(f"Invalid JSON payload: {exc}") from exc
    result = dispatch_command(
        command=SYNTHESIS_COMMAND,
        payload=payload,
        root=None,
        runner=runner,
    )
    output = json.dumps(result, indent=2, sort_keys=True)
    if output_path is None:
        typer.echo(output)
    else:
        output_path.write_text(output)


@app.command("refactor-protocol")
def refactor_protocol(
    input_path: Optional[Path] = typer.Option(
        None, "--input", help="JSON payload describing the refactor request."
    ),
    output_path: Optional[Path] = typer.Option(
        None, "--output", help="Write refactor response JSON to this path."
    ),
    protocol_name: Optional[str] = typer.Option(None, "--protocol-name"),
    bundle: Optional[List[str]] = typer.Option(None, "--bundle"),
    field: Optional[List[str]] = typer.Option(
        None,
        "--field",
        help="Field spec in 'name:type' form (repeatable).",
    ),
    target_path: Optional[Path] = typer.Option(None, "--target-path"),
    target_functions: Optional[List[str]] = typer.Option(None, "--target-function"),
    compatibility_shim: bool = typer.Option(
        False, "--compat-shim/--no-compat-shim"
    ),
    rationale: Optional[str] = typer.Option(None, "--rationale"),
) -> None:
    """Generate protocol refactor edits from a JSON payload (prototype)."""
    input_payload: dict[str, Any] | None = None
    if input_path is not None:
        try:
            input_payload = json.loads(input_path.read_text())
        except json.JSONDecodeError as exc:
            raise typer.BadParameter(f"Invalid JSON payload: {exc}") from exc
    payload = build_refactor_payload(
        input_payload=input_payload,
        protocol_name=protocol_name,
        bundle=bundle,
        field=field,
        target_path=target_path,
        target_functions=target_functions,
        compatibility_shim=compatibility_shim,
        rationale=rationale,
    )
    result = dispatch_command(
        command=REFACTOR_COMMAND,
        payload=payload,
        root=None,
    )
    output = json.dumps(result, indent=2, sort_keys=True)
    if output_path is None:
        typer.echo(output)
    else:
        output_path.write_text(output)



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _write(tmp_path: Path, rel: str, content: str) -> Path:
    path = tmp_path / rel
    path.parent.mkdir(parents=True, exist_ok=True)
    path.write_text(textwrap.dedent(content))
    return path


def _load_analyzer():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_unused_arg_flow_repo

    return analyze_unused_arg_flow_repo


def test_unused_arg_smell_detected(tmp_path: Path) -> None:
    path = _write(
        tmp_path,
        "sample.py",
        """
        def callee(x, y):
            return y

        def caller(foo):
            return callee(foo, 1)
        """,
    )
    analyze_unused_arg_flow_repo = _load_analyzer()
    smells = analyze_unused_arg_flow_repo(
        [path],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        external_filter=True,
    )
    assert (
        "sample.py:caller passes param foo to unused sample.py:callee.x"
        in smells
    )


def test_unused_arg_ignores_constants(tmp_path: Path) -> None:
    path = _write(
        tmp_path,
        "sample.py",
        """
        def callee(x, y):
            return y

        def caller(foo):
            return callee(1, foo)
        """,
    )
    analyze_unused_arg_flow_repo = _load_analyzer()
    smells = analyze_unused_arg_flow_repo(
        [path],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        external_filter=True,
    )
    assert not smells


def test_unused_arg_skips_test_calls(tmp_path: Path) -> None:
    path = _write(
        tmp_path,
        "tests/test_sample.py",
        """
        def callee(x, y):
            return y

        def caller(foo):
            return callee(foo, 1)
        """,
    )
    analyze_unused_arg_flow_repo = _load_analyzer()
    smells = analyze_unused_arg_flow_repo(
        [path],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        external_filter=True,
    )
    assert not smells


def test_unused_arg_non_const_pos_and_kw(tmp_path: Path) -> None:
    path = _write(
        tmp_path,
        "sample.py",
        """
        def callee(x, y, z):
            return y

        def caller(a, b):
            return callee(a + 1, y=1, z=b + 2)
        """,
    )
    analyze_unused_arg_flow_repo = _load_analyzer()
    smells = analyze_unused_arg_flow_repo(
        [path],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        external_filter=True,
    )
    assert any("non-constant arg at position 0" in smell for smell in smells)
    assert any("non-constant kw 'z'" in smell for smell in smells)


def test_unused_arg_star_kw_low_strictness(tmp_path: Path) -> None:
    path = _write(
        tmp_path,
        "sample.py",
        """
        def callee(x, y):
            return y

        def caller(**kwargs):
            return callee(**kwargs)
        """,
    )
    analyze_unused_arg_flow_repo = _load_analyzer()
    smells = analyze_unused_arg_flow_repo(
        [path],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="low",
        external_filter=True,
    )
    assert any("non-constant kw 'x'" in smell for smell in smells)


def test_unused_arg_ignores_extra_pos_args(tmp_path: Path) -> None:
    path = _write(
        tmp_path,
        "sample.py",
        """
        def callee(x):
            return x

        def caller(a, b):
            return callee(a, b)
        """,
    )
    analyze_unused_arg_flow_repo = _load_analyzer()
    smells = analyze_unused_arg_flow_repo(
        [path],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        external_filter=True,
    )
    assert smells == []


def test_unused_arg_edge_branches(tmp_path: Path) -> None:
    path = _write(
        tmp_path,
        "sample.py",
        """
        def callee(x, y):
            return y

        def caller(a, b, c):
            callee(a, b, c)
            callee(a, b, c + 1)
            callee(x=a, y=b)
            callee(extra=a, x=b)
            callee(x=a, extra=b + 1)
        """,
    )
    analyze_unused_arg_flow_repo = _load_analyzer()
    smells = analyze_unused_arg_flow_repo(
        [path],
        project_root=tmp_path,
        ignore_params=set(),
        strictness="high",
        external_filter=True,
    )
    assert any("callee.x" in smell for smell in smells)



from __future__ import annotations

import ast
from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def test_const_repr_and_type_from_const_repr() -> None:
    da = _load()
    const_node = ast.parse("x = 1").body[0].value
    assert da._const_repr(const_node) == "1"
    unary_node = ast.parse("x = -1").body[0].value
    assert da._const_repr(unary_node) in {"-1", "(-1)"}
    attr_node = ast.parse("MOD.CONST").body[0].value
    assert da._const_repr(attr_node) == "MOD.CONST"
    lower_attr = ast.parse("mod.const").body[0].value
    assert da._const_repr(lower_attr) is None
    class _BadRepr:
        def __repr__(self) -> str:
            raise ValueError("bad repr")

    weird_unary = ast.UnaryOp(op=ast.UAdd(), operand=ast.Constant(value=_BadRepr()))
    assert da._const_repr(weird_unary) is None
    weird_attr = ast.Attribute(
        value=ast.Constant(value=_BadRepr()),
        attr="CONST",
        ctx=ast.Load(),
    )
    assert da._const_repr(weird_attr) is None

    assert da._type_from_const_repr("None") == "None"
    assert da._type_from_const_repr("True") == "bool"
    assert da._type_from_const_repr("1") == "int"
    assert da._type_from_const_repr("1.5") == "float"
    assert da._type_from_const_repr("1+2j") == "complex"
    assert da._type_from_const_repr("'hi'") == "str"
    assert da._type_from_const_repr("b'hi'") == "bytes"
    assert da._type_from_const_repr("[1, 2]") == "list"
    assert da._type_from_const_repr("(1, 2)") == "tuple"
    assert da._type_from_const_repr("{1, 2}") == "set"
    assert da._type_from_const_repr("{'a': 1}") == "dict"
    assert da._type_from_const_repr("...") is None
    assert da._type_from_const_repr("not a literal") is None


def test_split_and_combine_type_hints() -> None:
    da = _load()
    assert da._split_top_level("A,B[C,D],E", ",") == ["A", "B[C,D]", "E"]
    assert da._expand_type_hint("") == set()
    assert da._expand_type_hint("Optional[int]") == {"int", "None"}
    assert da._expand_type_hint("Union[int, str]") == {"int", "str"}
    assert da._expand_type_hint("int|None") == {"int", "None"}

    assert da._combine_type_hints({"int"}) == ("int", False)
    assert da._combine_type_hints({"Optional[int]"}) == ("Optional[int]", False)
    union_hint, conflicted = da._combine_type_hints({"int", "str"})
    assert union_hint == "Union[int, str]"
    assert conflicted is True
    any_hint, any_conflicted = da._combine_type_hints({"None"})
    assert any_hint == "Any"
    assert any_conflicted is True
    optional_union, conflict = da._combine_type_hints({"Optional[int]", "Optional[str]"})
    assert optional_union == "Optional[Union[int, str]]"
    assert conflict is True


def test_module_name_and_strings() -> None:
    da = _load()
    project_root = Path("repo")
    path = project_root / "src" / "pkg" / "mod.py"
    assert da._module_name(path, project_root) == "pkg.mod"
    assert da._module_name(Path("other/mod.py"), project_root) == "other.mod"

    tree = ast.parse("__all__ = ['Foo', 'bar']")
    assert da._string_list(tree.body[0].value) == ["Foo", "bar"]
    tree = ast.parse("__all__ = ['Foo', 1]")
    assert da._string_list(tree.body[0].value) is None


def test_base_identifier_and_exports() -> None:
    da = _load()
    name_node = ast.parse("x").body[0].value
    assert da._base_identifier(name_node) == "x"
    attr_node = ast.parse("mod.attr").body[0].value
    assert da._base_identifier(attr_node) == "mod.attr"
    sub_node = ast.parse("items[0]").body[0].value
    assert da._base_identifier(sub_node) == "items"
    call_node = ast.parse("factory().build").body[0].value
    assert da._base_identifier(call_node) == "factory().build"
    call_only = ast.parse("factory()").body[0].value
    assert da._base_identifier(call_only) == "factory"
    const_node = ast.parse("1").body[0].value
    assert da._base_identifier(const_node) is None

    module = ast.parse(
        "__all__ = ['Foo', 'bar']\n"
        "def Foo():\n"
        "    return 1\n"
        "def _hidden():\n"
        "    return 2\n"
        "value = 3\n"
    )
    export_names, export_map = da._collect_module_exports(
        module,
        module_name="demo",
        import_map={"bar": "other.bar"},
    )
    assert export_names == {"Foo", "bar"}
    assert export_map["Foo"] == "demo.Foo"
    assert export_map["bar"] == "other.bar"

    module = ast.parse(
        "__all__: list[str] = ['A']\n"
        "__all__ += ['B']\n"
        "A = 1\n"
        "B = 2\n"
    )
    export_names, export_map = da._collect_module_exports(
        module,
        module_name="demo",
        import_map={},
    )
    assert export_names == {"A", "B"}
    assert export_map["A"] == "demo.A"


def test_is_test_path() -> None:
    da = _load()
    assert da._is_test_path(Path("tests/test_sample.py")) is True
    assert da._is_test_path(Path("src/sample_test.py")) is False
    assert da._is_test_path(Path("src/test_module.py")) is True


def test_symbol_table_resolution_and_call_context() -> None:
    da = _load()
    table = da.SymbolTable(external_filter=True)
    table.imports[("pkg.mod", "Ext")] = "external.lib.Ext"
    table.imports[("pkg.mod", "Local")] = "pkg.local.Local"
    table.internal_roots.add("pkg")
    assert table.resolve("pkg.mod", "Local") == "pkg.local.Local"
    assert table.resolve("pkg.mod", "Ext") is None
    assert table.resolve("pkg.mod", "Name") == "pkg.mod.Name"

    table.external_filter = False
    assert table.resolve("pkg.mod", "Ext") == "external.lib.Ext"

    table.external_filter = True
    table.star_imports["pkg.mod"] = {"pkg.star", "ext.star"}
    table.module_exports["pkg.star"] = {"Foo"}
    table.module_export_map["pkg.star"] = {"Foo": "pkg.star.Foo"}
    table.module_exports["ext.star"] = {"Foo"}
    table.module_export_map["ext.star"] = {"Foo": "ext.star.Foo"}
    table.internal_roots.add("pkg")
    assert table.resolve_star("pkg.mod", "Foo") == "pkg.star.Foo"
    assert table.resolve_star("pkg.mod", "Missing") is None

    table.internal_roots.clear()
    assert table.resolve_star("pkg.mod", "Foo") is None

    table.external_filter = False
    assert table.resolve_star("pkg.mod", "Foo") == "ext.star.Foo"

    table.external_filter = True
    table.internal_roots.add("pkg")
    table.module_export_map["pkg.star"] = {}
    assert table.resolve_star("pkg.mod", "Foo") == "pkg.star.Foo"
    table.star_imports["pkg.mod"] = {"ext.star"}
    table.module_exports["ext.star"] = {"Foo"}
    table.module_export_map["ext.star"] = {}
    table.internal_roots = {"pkg"}
    assert table.resolve_star("pkg.mod", "Foo") is None
    table.star_imports["pkg.mod"] = {""}
    table.module_exports[""] = {"Bar"}
    assert table.resolve_star("pkg.mod", "Bar") == "Bar"

    tree = ast.parse("result = func(x)\n")
    parents = {}
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            parents[child] = node
    name_node = next(node for node in ast.walk(tree) if isinstance(node, ast.Name) and node.id == "func")
    call_node, direct = da._call_context(name_node, parents)
    assert isinstance(call_node, ast.Call)
    assert direct is False


def test_callee_normalization_and_decorators(tmp_path: Path) -> None:
    da = _load()
    assert da._normalize_callee("self.run", "Service") == "Service.run"
    assert da._normalize_callee("cls.build", "Factory") == "Factory.build"
    assert da._normalize_callee("other", "Service") == "other"

    bad_call = ast.Call(func=ast.Name(id=None, ctx=ast.Load()), args=[], keywords=[])
    assert da._callee_name(bad_call) == "<call>"

    assert da._decorator_name(ast.Name(id="deco", ctx=ast.Load())) == "deco"
    attr = ast.Attribute(value=ast.Name(id="pkg", ctx=ast.Load()), attr="wrap")
    assert da._decorator_name(attr) == "pkg.wrap"
    call = ast.Call(func=attr, args=[], keywords=[])
    assert da._decorator_name(call) == "pkg.wrap"
    assert da._decorator_name(ast.Constant(value=1)) is None
    assert da._decorator_matches("pkg.decor", {"decor"}) is True
    assert da._decorator_matches("decor", {"decor"}) is True
    assert da._decorator_matches("other", {"decor"}) is False

    tree = ast.parse(
        "@deco\n"
        "def f():\n"
        "    return 1\n"
        "\n"
        "@other\n"
        "def g():\n"
        "    return 2\n"
    )
    funcs = [node for node in tree.body if isinstance(node, ast.FunctionDef)]
    assert da._decorators_transparent(funcs[0], {"deco"}) is True
    assert da._decorators_transparent(funcs[1], {"deco"}) is False
    assert da._decorators_transparent(funcs[0], None) is True
    tree = ast.parse("@(lambda x: x)\ndef h():\n    return 3\n")
    func = tree.body[0]
    assert da._decorators_transparent(func, {"lambda"}) is False

    root = tmp_path / "root"
    root.mkdir()
    (root / "keep.py").write_text("x = 1\n")
    skip = root / "skip"
    skip.mkdir()
    (skip / "ignored.py").write_text("x = 2\n")
    config = da.AuditConfig(exclude_dirs={"skip"})
    paths = da._iter_paths([str(root)], config)
    assert root / "keep.py" in paths
    assert skip / "ignored.py" not in paths


def test_param_helpers_and_scopes() -> None:
    da = _load()
    tree = ast.parse(
        "class C:\n"
        "    def method(self, x: int, *args, **kw):\n"
        "        def inner(y):\n"
        "            return y\n"
        "        return inner(x)\n"
    )
    class_node = tree.body[0]
    fn = class_node.body[0]
    inner = fn.body[0]

    assert da._param_names(fn) == ["x", "args", "kw"]
    assert da._param_names(fn, {"x"}) == ["args", "kw"]
    annots = da._param_annotations(fn)
    assert annots["x"] == "int"
    assert annots["args"] is None
    broken_tree = ast.parse("def broken(bad):\n    pass\n")
    broken_fn = broken_tree.body[0]
    broken_fn.args.args[0].annotation = ast.Name(id=None, ctx=ast.Load())
    annots = da._param_annotations(broken_fn)
    assert annots["bad"] is None

    parents = {}
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            parents[child] = node
    assert da._enclosing_class(fn, parents) == "C"
    assert da._enclosing_class(inner, parents) == "C"
    assert da._enclosing_scopes(inner, parents) == ["C", "method"]
    assert da._enclosing_class_scopes(inner, parents) == ["C"]
    assert da._enclosing_function_scopes(inner, parents) == ["method"]
    assert da._function_key(["C", "method"], "inner") == "C.method.inner"

    arg = ast.arg(arg="z")
    arg.lineno = 1
    arg.col_offset = 0
    arg.end_lineno = 1
    arg.end_col_offset = 0
    assert da._node_span(arg) == (0, 0, 0, 1)
    assert da._node_span(ast.AST()) is None


def test_param_spans_ignore_var_kw() -> None:
    da = _load()
    tree = ast.parse("def f(a, *args, **kw):\n    return a\n")
    fn = tree.body[0]
    spans = da._param_spans(fn, {"a", "args", "kw"})
    assert spans == {}


def test_audit_config_ignored_paths() -> None:
    da = _load()
    config = da.AuditConfig(exclude_dirs={"build", "dist"})
    assert config.is_ignored_path(Path("repo/build/main.py")) is True
    assert config.is_ignored_path(Path("repo/src/main.py")) is False


def test_local_class_helpers() -> None:
    da = _load()
    class_bases = {
        "Base": [],
        "Child": ["Base"],
        "Loop": ["Loop"],
        "Other": ["external.Missing"],
    }
    assert da._local_class_name("pkg.Base", class_bases) == "Base"
    assert da._local_class_name("Missing", class_bases) is None
    assert da._local_class_name("pkg.Missing", class_bases) is None
    assert (
        da._resolve_local_method_in_hierarchy(
            "Child",
            "run",
            class_bases=class_bases,
            local_functions={"Base.run"},
            seen=set(),
        )
        == "Base.run"
    )
    assert (
        da._resolve_local_method_in_hierarchy(
            "Loop",
            "run",
            class_bases=class_bases,
            local_functions=set(),
            seen={"Loop"},
        )
        is None
    )
    assert (
        da._resolve_local_method_in_hierarchy(
            "Other",
            "run",
            class_bases=class_bases,
            local_functions=set(),
            seen=set(),
        )
        is None
    )


def test_collect_local_class_bases_skips_unknown() -> None:
    da = _load()
    tree = ast.parse(
        "class Base:\n"
        "    pass\n"
        "class Child(Base):\n"
        "    pass\n"
        "class Weird((lambda x: x)):\n"
        "    pass\n"
    )
    parents = {}
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            parents[child] = node
    bases = da._collect_local_class_bases(tree, parents)
    assert "Child" in bases
    assert bases["Child"] == ["Base"]
    assert "Weird" in bases
    assert bases["Weird"] == []


def test_param_spans_include_var_kw() -> None:
    da = _load()
    tree = ast.parse("def f(a, *args, **kw):\n    return a\n")
    fn = tree.body[0]
    spans = da._param_spans(fn)
    assert "a" in spans
    assert "args" in spans
    assert "kw" in spans


def test_resolve_star_external_filtered() -> None:
    da = _load()
    table = da.SymbolTable(external_filter=True)
    table.star_imports["pkg.mod"] = {"ext.star"}
    table.module_exports["ext.star"] = {"Foo"}
    table.module_export_map["ext.star"] = {}
    table.internal_roots.add("pkg")
    assert table.resolve_star("pkg.mod", "Foo") is None


def test_param_defaults_and_return_aliases() -> None:
    da = _load()
    tree = ast.parse(
        "class C:\n"
        "    def method(self, a, b=1, *, c=2, d=None):\n"
        "        def inner():\n"
        "            return 1\n"
        "        async def inner_async():\n"
        "            return 2\n"
        "        _ = (lambda z: z)\n"
        "        return a\n"
    )
    fn = tree.body[0].body[0]
    defaults = da._param_defaults(fn)
    assert defaults == {"b", "c", "d"}
    defaults_filtered = da._param_defaults(fn, {"b"})
    assert defaults_filtered == {"c", "d"}
    alias = da._return_aliases(fn)
    assert alias == ["a"]


def test_return_aliases_tuple_and_conflict() -> None:
    da = _load()
    tree = ast.parse(
        "def tuple_alias(a, b):\n"
        "    return (a, b)\n"
        "\n"
        "def conflict(a, b):\n"
        "    if a:\n"
        "        return a\n"
        "    return b\n"
    )
    tuple_fn = tree.body[0]
    conflict_fn = tree.body[1]
    assert da._return_aliases(tuple_fn) == ["a", "b"]
    assert da._return_aliases(conflict_fn) is None


def test_collect_return_aliases_conflict() -> None:
    da = _load()
    tree = ast.parse(
        "def foo(a):\n"
        "    return a\n"
        "\n"
        "def foo(b):\n"
        "    return b\n"
        "\n"
        "def foo(c):\n"
        "    return c\n"
    )
    funcs = [node for node in ast.walk(tree) if isinstance(node, ast.FunctionDef)]
    parents = {}
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            parents[child] = node
    aliases = da._collect_return_aliases(funcs, parents, ignore_params=None)
    assert "foo" not in aliases


def test_return_aliases_bare_return() -> None:
    da = _load()
    tree = ast.parse(
        "def f(a):\n"
        "    return\n"
    )
    fn = tree.body[0]
    assert isinstance(fn, ast.FunctionDef)
    assert da._return_aliases(fn, ignore_params=None) is None



from __future__ import annotations

from dataclasses import dataclass
from pathlib import Path

from gabion import server


class _DummyWorkspace:
    def __init__(self, root_path: str) -> None:
        self.root_path = root_path


class _DummyServer:
    def __init__(self, root_path: str) -> None:
        self.workspace = _DummyWorkspace(root_path)


@dataclass
class _CommandResult:
    exit_code: int
    violations: int


def _write_bundle_module(path: Path) -> None:
    path.write_text(
        "def callee(x):\n"
        "    return x\n"
        "\n"
        "def caller(a, b):\n"
        "    callee(a)\n"
        "    callee(b)\n"
    )


def _write_type_conflict_module(path: Path) -> None:
    path.write_text(
        "def callee_int(x: int):\n"
        "    return x\n"
        "\n"
        "def callee_str(x: str):\n"
        "    return x\n"
        "\n"
        "def caller_conflict(b):\n"
        "    callee_int(b)\n"
        "    callee_str(b)\n"
    )


def test_execute_command_dash_outputs(tmp_path: Path) -> None:
    module_path = tmp_path / "sample.py"
    _write_bundle_module(module_path)
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "dot": "-",
            "synthesis_plan": "-",
            "synthesis_protocols": "-",
            "refactor_plan_json": "-",
        },
    )
    assert "dot" in result
    assert "synthesis_plan" in result
    assert "synthesis_protocols" in result
    assert "refactor_plan" in result


def test_execute_command_baseline_apply(tmp_path: Path) -> None:
    module_path = tmp_path / "sample.py"
    _write_bundle_module(module_path)
    baseline_path = tmp_path / "baseline.txt"
    baseline_path.write_text("preexisting\n")
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "baseline": str(baseline_path),
            "fail_on_violations": True,
        },
    )
    assert result.get("baseline_written") is False
    assert _CommandResult(
        exit_code=result.get("exit_code", -1),
        violations=result.get("violations", -1),
    ) == _CommandResult(exit_code=1, violations=result.get("violations", -1))


def test_execute_command_fail_on_type_ambiguities(tmp_path: Path) -> None:
    module_path = tmp_path / "types.py"
    _write_type_conflict_module(module_path)
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "fail_on_type_ambiguities": True,
        },
    )
    assert result.get("exit_code") == 1
    assert result.get("type_ambiguities")


def test_execute_command_report_baseline_write(tmp_path: Path) -> None:
    module_path = tmp_path / "sample.py"
    _write_bundle_module(module_path)
    report_path = tmp_path / "report.md"
    baseline_path = tmp_path / "baseline.txt"
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "report": str(report_path),
            "baseline": str(baseline_path),
            "baseline_write": True,
            "fail_on_violations": True,
        },
    )
    assert result.get("baseline_written") is True
    assert result.get("exit_code") == 0
    assert report_path.exists()
    assert baseline_path.exists()


def test_execute_command_report_appends_sections(tmp_path: Path) -> None:
    module_path = tmp_path / "sample.py"
    _write_bundle_module(module_path)
    report_path = tmp_path / "report.md"
    baseline_path = tmp_path / "baseline.txt"
    baseline_path.write_text("legacy\n")
    refactor_json = tmp_path / "refactor.json"
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(
        ls,
        {
            "root": str(tmp_path),
            "paths": [str(module_path)],
            "report": str(report_path),
            "baseline": str(baseline_path),
            "synthesis_report": True,
            "refactor_plan": True,
            "refactor_plan_json": str(refactor_json),
        },
    )
    assert result.get("baseline_written") is False
    assert report_path.exists()
    report_text = report_path.read_text()
    assert "Synthesis plan" in report_text
    assert "Refactoring plan" in report_text
    assert refactor_json.exists()


def test_execute_command_defaults_payload(tmp_path: Path) -> None:
    ls = _DummyServer(str(tmp_path))
    result = server.execute_command(ls, None)
    assert "violations" in result
    assert "exit_code" in result


def test_execute_refactor_valid_payload(tmp_path: Path) -> None:
    module_path = tmp_path / "target.py"
    module_path.write_text("def f(a, b):\n    return a + b\n")
    ls = _DummyServer(str(tmp_path))
    result = server.execute_refactor(
        ls,
        {
            "protocol_name": "ExampleProto",
            "bundle": ["a", "b"],
            "target_path": str(module_path),
            "target_functions": [],
        },
    )
    assert result.get("errors") == []
    edits = result.get("edits", [])
    assert edits


def test_execute_refactor_invalid_payload(tmp_path: Path) -> None:
    ls = _DummyServer(str(tmp_path))
    result = server.execute_refactor(ls, {"protocol_name": 123})
    assert result.get("errors")


def test_execute_refactor_payload_none(tmp_path: Path) -> None:
    ls = _DummyServer(str(tmp_path))
    result = server.execute_refactor(ls, None)
    assert result.get("errors")


def test_execute_synthesis_invalid_payload(tmp_path: Path) -> None:
    ls = _DummyServer(str(tmp_path))
    result = server.execute_synthesis(ls, {"bundles": "not-a-list"})
    assert result.get("errors")


def test_execute_synthesis_records_bundle_tiers(tmp_path: Path) -> None:
    ls = _DummyServer(str(tmp_path))
    result = server.execute_synthesis(
        ls,
        {
            "bundles": [
                {"bundle": ["a", "b"], "tier": 2},
            ],
            "existing_names": [],
        },
    )
    assert result.get("errors") == []
    assert "protocols" in result


def test_execute_synthesis_payload_none(tmp_path: Path) -> None:
    ls = _DummyServer(str(tmp_path))
    result = server.execute_synthesis(ls, None)
    assert result.get("errors")


def test_execute_synthesis_skips_empty_bundle(tmp_path: Path) -> None:
    ls = _DummyServer(str(tmp_path))
    result = server.execute_synthesis(
        ls,
        {
            "bundles": [{"bundle": [], "tier": 2}],
        },
    )
    assert result.get("protocols") == []



"""Synthesis subpackage for Gabion."""

from gabion.synthesis.merge import merge_bundles
from gabion.synthesis.model import (
    FieldSpec,
    NamingContext,
    ProtocolSpec,
    SynthesisConfig,
    SynthesisPlan,
)
from gabion.synthesis.naming import suggest_name
from gabion.synthesis.protocols import Synthesizer
from gabion.synthesis.schedule import ScheduleResult, topological_schedule

__all__ = [
    "FieldSpec",
    "NamingContext",
    "ProtocolSpec",
    "ScheduleResult",
    "SynthesisConfig",
    "SynthesisPlan",
    "Synthesizer",
    "merge_bundles",
    "suggest_name",
    "topological_schedule",
]



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def test_flow_analyses_edge_cases(tmp_path: Path) -> None:
    da = _load()
    module_path = tmp_path / "mod.py"
    module_path.write_text(
        textwrap.dedent(
            """
            def short(a):
                return a

            def short_unused(a, b):
                return a

            def typed_int(x: int):
                return x

            def typed_str(x: str):
                return x

            def typed_only(u):
                return typed_int(u)

            def const_only(a):
                return a

            def with_default(a, b=1):
                return a

            def star_callee(a, b, c):
                return a

            def bundle_caller(p, q):
                short(p)
                short(q)

            def caller(x, y, *args, **kwargs):
                short(1, x)
                short(1, 2)
                short(x, x + 1)
                short(a=1, z=2)
                short(a=x, z=y)
                short(z=x + 1)
                short(*args, **kwargs)
                short_unused(x, y)
                short_unused(x, b=2)
                typed_int(x)
                typed_str(x)
                const_only(1)
                with_default(x)
                star_callee(*args, **kwargs)
            """
        ).strip()
        + "\n"
    )
    test_path = tmp_path / "tests" / "test_mod.py"
    test_path.parent.mkdir(parents=True)
    test_path.write_text(
        "def test_dummy(a):\n"
        "    short(a)\n"
    )
    paths = [module_path, test_path]
    config = da.AuditConfig(
        project_root=tmp_path,
        exclude_dirs=set(),
        ignore_params=set(),
        external_filter=True,
        strictness="low",
        transparent_decorators=None,
    )
    analysis = da.analyze_paths(
        paths,
        recursive=True,
        type_audit=True,
        type_audit_report=True,
        type_audit_max=5,
        include_constant_smells=True,
        include_unused_arg_smells=True,
        config=config,
    )
    assert analysis.type_ambiguities
    assert analysis.constant_smells is not None
    assert analysis.unused_arg_smells is not None

    plan = da.build_synthesis_plan(
        analysis.groups_by_path,
        project_root=tmp_path,
        max_tier=2,
        min_bundle_size=2,
        allow_singletons=False,
        merge_overlap_threshold=0.5,
        config=config,
    )
    assert "protocols" in plan

    suggestions, ambiguities = da.analyze_type_flow_repo(
        paths,
        project_root=tmp_path,
        ignore_params=set(),
        strictness="low",
        external_filter=True,
        transparent_decorators=None,
    )
    assert suggestions
    assert ambiguities

    const_smells = da.analyze_constant_flow_repo(
        paths,
        project_root=tmp_path,
        ignore_params=set(),
        strictness="low",
        external_filter=True,
        transparent_decorators=None,
    )
    assert const_smells

    unused_smells = da.analyze_unused_arg_flow_repo(
        paths,
        project_root=tmp_path,
        ignore_params=set(),
        strictness="low",
        external_filter=True,
        transparent_decorators=None,
    )
    assert unused_smells


def test_analyze_paths_config_default(tmp_path: Path) -> None:
    da = _load()
    module_path = tmp_path / "mod.py"
    module_path.write_text("def f(a, b):\n    return a\n")
    analysis = da.analyze_paths(
        [module_path],
        recursive=False,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=None,
    )
    assert analysis.groups_by_path



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Dict, List, Set


@dataclass(frozen=True)
class ScheduleResult:
    order: List[str] = field(default_factory=list)
    cycles: List[Set[str]] = field(default_factory=list)


def topological_schedule(graph: Dict[str, Set[str]]) -> ScheduleResult:
    nodes: Set[str] = set(graph.keys())
    for deps in graph.values():
        nodes.update(deps)

    incoming: Dict[str, Set[str]] = {node: set() for node in nodes}
    outgoing: Dict[str, Set[str]] = {node: set() for node in nodes}

    for node, deps in graph.items():
        for dep in deps:
            outgoing[dep].add(node)
            incoming[node].add(dep)

    ready = sorted(node for node, deps in incoming.items() if not deps)
    order: List[str] = []

    while ready:
        node = ready.pop(0)
        order.append(node)
        for follower in sorted(outgoing[node]):
            incoming[follower].discard(node)
            if not incoming[follower]:
                if follower not in ready and follower not in order:
                    ready.append(follower)
        ready.sort()

    remaining = {node for node, deps in incoming.items() if deps}
    cycles: List[Set[str]] = []
    if remaining:
        subgraph = {node: {dep for dep in graph.get(node, set()) if dep in remaining} for node in remaining}
        cycles = _strongly_connected_components(subgraph)
        cycles = [
            comp
            for comp in cycles
            if len(comp) > 1 or any(node in subgraph.get(node, set()) for node in comp)
        ]
    return ScheduleResult(order=order, cycles=cycles)


def _strongly_connected_components(graph: Dict[str, Set[str]]) -> List[Set[str]]:
    index = 0
    indices: Dict[str, int] = {}
    lowlinks: Dict[str, int] = {}
    stack: List[str] = []
    on_stack: Set[str] = set()
    components: List[Set[str]] = []

    def visit(node: str) -> None:
        nonlocal index
        indices[node] = index
        lowlinks[node] = index
        index += 1
        stack.append(node)
        on_stack.add(node)
        for neighbor in graph.get(node, set()):
            if neighbor not in indices:
                visit(neighbor)
                lowlinks[node] = min(lowlinks[node], lowlinks[neighbor])
            elif neighbor in on_stack:
                lowlinks[node] = min(lowlinks[node], indices[neighbor])
        if lowlinks[node] == indices[node]:
            component: Set[str] = set()
            while True:
                popped = stack.pop()
                on_stack.discard(popped)
                component.add(popped)
                if popped == node:
                    break
            components.append(component)

    for node in graph:
        if node not in indices:
            visit(node)

    return components



from __future__ import annotations

from pathlib import Path
import sys
import ast


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def test_resolve_class_candidates_variants() -> None:
    da = _load()
    symbol_table = da.SymbolTable()
    symbol_table.internal_roots.add("pkg")
    symbol_table.imports[("pkg.mod", "ext")] = "pkg.ext"
    class_index = {
        "pkg.ext.Base": da.ClassInfo(qual="pkg.ext.Base", module="pkg.ext", bases=[], methods=set()),
        "pkg.mod.Base": da.ClassInfo(qual="pkg.mod.Base", module="pkg.mod", bases=[], methods=set()),
    }
    dotted = da._resolve_class_candidates(
        "ext.Base",
        module="pkg.mod",
        symbol_table=symbol_table,
        class_index=class_index,
    )
    assert "pkg.ext.Base" in dotted

    symbol_table.imports[("pkg.mod", "Base")] = "pkg.mod.Base"
    symbol_table.star_imports["pkg.mod"] = {"pkg.mod"}
    symbol_table.module_exports["pkg.mod"] = {"Base"}
    bare = da._resolve_class_candidates(
        "Base",
        module="pkg.mod",
        symbol_table=symbol_table,
        class_index=class_index,
    )
    assert "pkg.mod.Base" in bare


def test_resolve_method_in_hierarchy() -> None:
    da = _load()
    base_info = da.FunctionInfo(
        name="run",
        qual="pkg.Base.run",
        path=Path("base.py"),
        params=[],
        annots={},
        calls=[],
        unused_params=set(),
    )
    class_index = {
        "pkg.Base": da.ClassInfo(qual="pkg.Base", module="pkg", bases=[], methods={"run"}),
        "pkg.Child": da.ClassInfo(qual="pkg.Child", module="pkg", bases=["Base"], methods=set()),
    }
    by_qual = {"pkg.Base.run": base_info}
    symbol_table = da.SymbolTable()
    symbol_table.imports[("pkg", "Base")] = "pkg.Base"
    symbol_table.internal_roots.add("pkg")
    resolved = da._resolve_method_in_hierarchy(
        "pkg.Child",
        "run",
        class_index=class_index,
        by_qual=by_qual,
        symbol_table=symbol_table,
        seen=set(),
    )
    assert resolved is base_info


def test_local_class_bases_and_method_resolution() -> None:
    da = _load()
    tree = ast.parse(
        "class Base:\n"
        "    def run(self):\n"
        "        return 1\n"
        "\n"
        "class Child(Base):\n"
        "    def other(self):\n"
        "        return 2\n"
    )
    parents = {}
    for node in ast.walk(tree):
        for child in ast.iter_child_nodes(node):
            parents[child] = node
    class_bases = da._collect_local_class_bases(tree, parents)
    assert class_bases["Base"] == []
    assert class_bases["Child"] == ["Base"]
    resolved = da._resolve_local_method_in_hierarchy(
        "Child",
        "run",
        class_bases=class_bases,
        local_functions={"Base.run"},
        seen=set(),
    )
    assert resolved == "Base.run"



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit

    return dataflow_audit


def _write_conflict_code(path: Path) -> None:
    path.write_text(
        "def callee_int(x: int):\n"
        "    return x\n\n"
        "def callee_str(x: str):\n"
        "    return x\n\n"
        "def caller(a):\n"
        "    return callee_int(a)\n\n"
        "def caller_conflict(b):\n"
        "    callee_int(b)\n"
        "    callee_str(b)\n"
    )


def _write_bundle_code(path: Path) -> None:
    path.write_text(
        "def callee(x):\n"
        "    return x\n"
        "\n"
        "def caller(a, b):\n"
        "    callee(a)\n"
        "    callee(b)\n"
    )


def test_run_baseline_write_requires_path(tmp_path: Path) -> None:
    dataflow_audit = _load()
    code = dataflow_audit.run(
        [
            str(tmp_path),
            "--root",
            str(tmp_path),
            "--baseline-write",
        ]
    )
    assert code == 2


def test_run_dot_only_returns_early(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    sample.write_text("def f(a, b):\n    return a + b\n")
    dot_path = tmp_path / "graph.dot"
    code = dataflow_audit.run(
        [
            str(tmp_path),
            "--root",
            str(tmp_path),
            "--dot",
            str(dot_path),
        ]
    )
    assert code == 0
    assert dot_path.exists()


def test_run_fail_on_type_ambiguities_with_synthesis_plan(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "mod.py"
    _write_conflict_code(sample)
    plan_path = tmp_path / "plan.json"
    code = dataflow_audit.run(
        [
            str(tmp_path),
            "--root",
            str(tmp_path),
            "--synthesis-plan",
            str(plan_path),
            "--fail-on-type-ambiguities",
        ]
    )
    assert plan_path.exists()
    assert code == 1


def test_run_report_with_baseline_write(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    report_path = tmp_path / "report.md"
    baseline_path = tmp_path / "baseline.txt"
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--report",
            str(report_path),
            "--baseline",
            str(baseline_path),
            "--baseline-write",
            "--fail-on-violations",
        ]
    )
    assert code == 0
    assert report_path.exists()
    assert baseline_path.exists()


def test_run_report_with_baseline_apply(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    report_path = tmp_path / "report.md"
    baseline_path = tmp_path / "baseline.txt"
    baseline_path.write_text("preexisting\n")
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--report",
            str(report_path),
            "--baseline",
            str(baseline_path),
            "--fail-on-violations",
        ]
    )
    assert report_path.exists()
    assert code == 1


def test_run_type_audit_prints_findings(tmp_path: Path, capsys) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "types.py"
    _write_conflict_code(sample)
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--type-audit",
            "--type-audit-max",
            "5",
            "--transparent-decorators",
            "wrap, deco",
        ]
    )
    captured = capsys.readouterr().out
    assert "Type tightening candidates" in captured
    assert "Type ambiguities" in captured
    assert code == 0


def test_run_synthesis_protocols_only(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    stubs_path = tmp_path / "stubs.py"
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--synthesis-protocols",
            str(stubs_path),
            "--synthesis-merge-overlap",
            "1.5",
        ]
    )
    assert stubs_path.exists()
    assert code == 0


def test_run_synthesis_outputs_to_stdout(tmp_path: Path, capsys) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--synthesis-plan",
            "-",
            "--synthesis-protocols",
            "-",
        ]
    )
    captured = capsys.readouterr().out
    assert "protocol" in captured or "class" in captured
    assert code == 0


def test_run_synthesis_uses_config_overlap(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    config_path = tmp_path / "gabion.toml"
    config_path.write_text("[synthesis]\nmerge_overlap_threshold = 0.5\n")
    plan_path = tmp_path / "plan.json"
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--config",
            str(config_path),
            "--synthesis-plan",
            str(plan_path),
        ]
    )
    assert plan_path.exists()
    assert code == 0


def test_run_refactor_plan_json_stdout(tmp_path: Path, capsys) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--refactor-plan-json",
            "-",
        ]
    )
    captured = capsys.readouterr().out
    assert captured.strip()
    assert code == 0


def test_run_refactor_plan_without_json(tmp_path: Path, capsys) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--refactor-plan",
            "--exclude",
            "in, .venv",
            "--ignore-params",
            "x, y",
        ]
    )
    captured = capsys.readouterr().out
    assert "bundle" in captured
    assert code == 0


def test_run_dot_stdout_returns_early(tmp_path: Path, capsys) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--dot",
            "-",
        ]
    )
    captured = capsys.readouterr().out
    assert "graph" in captured or "digraph" in captured
    assert code == 0


def test_run_fail_on_violations_with_baseline(tmp_path: Path, capsys) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    baseline_path = tmp_path / "baseline.txt"
    baseline_path.write_text("preexisting\n")
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--baseline",
            str(baseline_path),
            "--fail-on-violations",
        ]
    )
    captured = capsys.readouterr().out
    assert "bundle" in captured
    assert code == 1


def test_run_fail_on_violations_baseline_write(tmp_path: Path) -> None:
    dataflow_audit = _load()
    sample = tmp_path / "sample.py"
    _write_bundle_code(sample)
    baseline_path = tmp_path / "baseline.txt"
    code = dataflow_audit.run(
        [
            str(sample),
            "--root",
            str(tmp_path),
            "--baseline",
            str(baseline_path),
            "--baseline-write",
            "--fail-on-violations",
        ]
    )
    assert baseline_path.exists()
    assert code == 0



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import AuditConfig, analyze_paths, build_refactor_plan
    from gabion.refactor.engine import RefactorEngine
    from gabion.refactor.model import FieldSpec, RefactorRequest

    return AuditConfig, analyze_paths, build_refactor_plan, RefactorEngine, FieldSpec, RefactorRequest


def _apply_plan(plan) -> None:
    for edit in plan.edits:
        Path(edit.path).write_text(edit.replacement)


def test_refactor_idempotency(tmp_path: Path) -> None:
    (
        AuditConfig,
        analyze_paths,
        build_refactor_plan,
        RefactorEngine,
        FieldSpec,
        RefactorRequest,
    ) = _load()
    source = textwrap.dedent(
        """
        def sink(x=None, y=None):
            return x, y

        def g(a, b):
            sink(x=a, y=a)
            sink(x=b, y=b)

        def f(a, b):
            return g(a, b)
        """
    )
    file_path = tmp_path / "mod.py"
    file_path.write_text(source.strip() + "\n")
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [file_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    plan = build_refactor_plan(analysis.groups_by_path, [file_path], config=config)
    bundles = plan.get("bundles", [])
    assert bundles
    bundle = bundles[0].get("bundle", [])
    assert bundle

    request = RefactorRequest(
        protocol_name="BundleProtocol",
        bundle=list(bundle),
        fields=[FieldSpec(name=name, type_hint="int") for name in bundle],
        target_path=str(file_path),
        target_functions=["g"],
        rationale="Idempotency test",
    )
    refactor_plan = RefactorEngine(project_root=tmp_path).plan_protocol_extraction(request)
    assert refactor_plan.edits
    _apply_plan(refactor_plan)

    analysis_after = analyze_paths(
        [file_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    plan_after = build_refactor_plan(analysis_after.groups_by_path, [file_path], config=config)
    assert not plan_after.get("bundles", [])



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import (
        _build_symbol_table,
        _collect_dataclass_registry,
        _iter_dataclass_call_bundles,
    )

    return _iter_dataclass_call_bundles, _build_symbol_table, _collect_dataclass_registry


def test_dataclass_call_bundles_accepts_expression_values(tmp_path: Path) -> None:
    _iter_dataclass_call_bundles, _, _ = _load()
    source = tmp_path / "example.py"
    source.write_text(
        """
from dataclasses import dataclass

@dataclass
class Payload:
    alpha: int
    beta: int
    gamma: int


def build(alpha, beta, gamma):
    return Payload(alpha, beta + 1, gamma=make_gamma())
"""
    )
    bundles = _iter_dataclass_call_bundles(source)
    assert ("alpha", "beta", "gamma") in bundles


def test_dataclass_call_bundles_resolve_cross_file(tmp_path: Path) -> None:
    _iter_dataclass_call_bundles, _build_symbol_table, _collect_dataclass_registry = _load()
    root = tmp_path
    (root / "models.py").write_text(
        """
from dataclasses import dataclass

@dataclass
class Payload:
    alpha: int
    beta: int
    gamma: int
"""
    )
    caller = root / "consumer.py"
    caller.write_text(
        """
from models import Payload

def build(alpha, beta, gamma):
    return Payload(alpha, beta + 1, gamma=make_gamma())
"""
    )
    paths = [root / "models.py", caller]
    symbol_table = _build_symbol_table(paths, root, external_filter=True)
    registry = _collect_dataclass_registry(paths, project_root=root)
    bundles = _iter_dataclass_call_bundles(
        caller,
        project_root=root,
        symbol_table=symbol_table,
        dataclass_registry=registry,
    )
    assert ("alpha", "beta", "gamma") in bundles



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.synthesis.schedule import topological_schedule

    return topological_schedule


def test_topological_schedule_orders_dependencies() -> None:
    topological_schedule = _load()
    graph = {"a": {"b"}, "b": set()}
    result = topological_schedule(graph)
    assert result.order == ["b", "a"]
    assert result.cycles == []


def test_topological_schedule_reports_cycles() -> None:
    topological_schedule = _load()
    graph = {"a": {"b"}, "b": {"a"}}
    result = topological_schedule(graph)
    assert result.cycles
    assert set().union(*result.cycles) == {"a", "b"}



from __future__ import annotations

from pathlib import Path
from typing import Any
import tomllib

DEFAULT_CONFIG_NAME = "gabion.toml"


def _load_toml(path: Path) -> dict[str, Any]:
    try:
        raw = path.read_text(encoding="utf-8")
    except FileNotFoundError:
        return {}
    except OSError:
        return {}
    try:
        data = tomllib.loads(raw)
    except Exception:
        return {}
    return data if isinstance(data, dict) else {}


def load_config(root: Path | None = None, config_path: Path | None = None) -> dict[str, Any]:
    if config_path is None:
        base = root if root is not None else Path.cwd()
        config_path = base / DEFAULT_CONFIG_NAME
    return _load_toml(config_path)


def dataflow_defaults(
    root: Path | None = None, config_path: Path | None = None
) -> dict[str, Any]:
    data = load_config(root=root, config_path=config_path)
    section = data.get("dataflow", {})
    return section if isinstance(section, dict) else {}


def synthesis_defaults(
    root: Path | None = None, config_path: Path | None = None
) -> dict[str, Any]:
    data = load_config(root=root, config_path=config_path)
    section = data.get("synthesis", {})
    return section if isinstance(section, dict) else {}


def merge_payload(payload: dict[str, Any], defaults: dict[str, Any]) -> dict[str, Any]:
    merged = dict(defaults)
    for key, value in payload.items():
        if value is None:
            continue
        merged[key] = value
    return merged



from __future__ import annotations

from pathlib import Path

import pytest
import typer

from gabion import cli


def test_check_builds_payload() -> None:
    payload = cli.build_check_payload(
        paths=None,
        report=None,
        fail_on_violations=True,
        root=Path("."),
        config=None,
        baseline=None,
        baseline_write=False,
        exclude=None,
        ignore_params=None,
        transparent_decorators=None,
        allow_external=None,
        strictness=None,
        fail_on_type_ambiguities=True,
    )
    assert payload["paths"] == ["."]
    assert payload["fail_on_violations"] is True
    assert payload["fail_on_type_ambiguities"] is True
    assert payload["type_audit"] is True


def test_check_payload_strictness_validation() -> None:
    with pytest.raises(typer.BadParameter):
        cli.build_check_payload(
            paths=[Path(".")],
            report=None,
            fail_on_violations=True,
            root=Path("."),
            config=None,
            baseline=None,
            baseline_write=False,
            exclude=None,
            ignore_params=None,
            transparent_decorators=None,
            allow_external=None,
            strictness="medium",
            fail_on_type_ambiguities=False,
        )


def test_check_payload_baseline_write_requires_baseline() -> None:
    payload = cli.build_check_payload(
        paths=[Path(".")],
        report=None,
        fail_on_violations=True,
        root=Path("."),
        config=None,
        baseline=None,
        baseline_write=True,
        exclude=None,
        ignore_params=None,
        transparent_decorators=None,
        allow_external=None,
        strictness=None,
        fail_on_type_ambiguities=False,
    )
    assert payload["baseline_write"] is None


def test_dataflow_audit_payload_parsing() -> None:
    opts = cli.parse_dataflow_args(
        [
            ".",
            "--strictness",
            "low",
            "--exclude",
            "a,b",
            "--ignore-params",
            "x,y",
            "--no-recursive",
            "--fail-on-violations",
        ]
    )
    payload = cli.build_dataflow_payload(opts)
    assert payload["paths"] == ["."]
    assert payload["strictness"] == "low"
    assert payload["exclude"] == ["a", "b"]
    assert payload["ignore_params"] == ["x", "y"]
    assert payload["no_recursive"] is True
    assert payload["fail_on_violations"] is True


def test_dataflow_payload_baseline_and_transparent() -> None:
    opts = cli.parse_dataflow_args(
        [
            ".",
            "--baseline",
            "baseline.txt",
            "--baseline-write",
            "--transparent-decorators",
            "foo,bar",
            "--fail-on-type-ambiguities",
        ]
    )
    payload = cli.build_dataflow_payload(opts)
    assert payload["baseline"] == "baseline.txt"
    assert payload["baseline_write"] is True
    assert payload["transparent_decorators"] == ["foo", "bar"]
    assert payload["fail_on_type_ambiguities"] is True


def test_refactor_protocol_payload(tmp_path: Path) -> None:
    payload = cli.build_refactor_payload(
        protocol_name="Bundle",
        bundle=["a", "b"],
        field=["a:int", "b:str"],
        target_path=tmp_path / "sample.py",
        target_functions=["alpha"],
        compatibility_shim=True,
        rationale="use bundle",
    )
    assert payload["protocol_name"] == "Bundle"
    assert payload["bundle"] == ["a", "b"]
    assert payload["target_functions"] == ["alpha"]
    assert payload["compatibility_shim"] is True


def test_refactor_payload_infers_bundle(tmp_path: Path) -> None:
    payload = cli.build_refactor_payload(
        protocol_name="Bundle",
        bundle=None,
        field=["a:int", "b:str"],
        target_path=tmp_path / "sample.py",
        target_functions=[],
        compatibility_shim=False,
        rationale=None,
    )
    assert payload["bundle"] == ["a", "b"]


def test_run_check_uses_runner_dispatch(tmp_path: Path) -> None:
    captured: dict[str, object] = {}

    def runner(request, *, root=None):
        captured["command"] = request.command
        captured["payload"] = request.arguments[0]
        captured["root"] = root
        return {"exit_code": 0}

    result = cli.run_check(
        paths=[tmp_path],
        report=None,
        fail_on_violations=True,
        root=tmp_path,
        config=None,
        baseline=None,
        baseline_write=False,
        exclude=None,
        ignore_params=None,
        transparent_decorators=None,
        allow_external=None,
        strictness=None,
        fail_on_type_ambiguities=True,
        runner=runner,
    )
    assert result["exit_code"] == 0
    assert captured["command"] == cli.DATAFLOW_COMMAND
    assert captured["payload"]["paths"] == [str(tmp_path)]
    assert captured["root"] == tmp_path



from __future__ import annotations

from pathlib import Path
import sys


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis import dataflow_audit as da

    return da


def _write(path: Path, content: str) -> None:
    path.write_text(content)


def test_emit_report_empty_groups() -> None:
    da = _load()
    report, violations = da._emit_report({}, 3)
    assert "No bundle components detected." in report
    assert violations == []


def test_emit_report_component_summary(tmp_path: Path) -> None:
    da = _load()
    path = tmp_path / "mod.py"
    extra_path = tmp_path / "other.py"
    _write(
        path,
        "from dataclasses import dataclass\n"
        "\n"
        "# dataflow-bundle: a, b\n"
        "@dataclass\n"
        "class AppConfig:\n"
        "    c: int\n"
        "    d: int\n"
        "\n"
        "def f(a, b, x, y):\n"
        "    return a\n",
    )
    _write(extra_path, "def g():\n    return 1\n")
    groups_by_path = {
        path: {"f": [set(["a", "b"]), set(["x", "y"])]},
        extra_path: {},
    }
    report, violations = da._emit_report(
        groups_by_path,
        3,
        type_suggestions=["f.a can tighten to int"],
        type_ambiguities=["f.b downstream types conflict: ['int', 'str']"],
        constant_smells=["mod.py:f.a only observed constant 1 across 1 non-test call(s)"],
        unused_arg_smells=["mod.py:f passes param x to unused mod.py:f.x"],
    )
    assert "Observed-only bundles" in report
    assert "Documented bundles" in report
    assert "Declared Config bundles not observed" in report
    assert "Type-flow audit" in report
    assert "Constant-propagation smells" in report
    assert "Unused-argument smells" in report
    assert any("tier-3" in line for line in violations)


def test_emit_report_max_components_cutoff(tmp_path: Path) -> None:
    da = _load()
    path = tmp_path / "mod.py"
    _write(path, "def f(a, b):\n    return a\n")
    groups_by_path = {path: {"f": [set(["a", "b"])]}}
    report, _ = da._emit_report(groups_by_path, 0)
    assert "Showing top 0 components" in report


def test_render_mermaid_component_declared_none_documented(tmp_path: Path) -> None:
    da = _load()
    path = tmp_path / "mod.py"
    fn_key = f"fn::{path}::f"
    bundle_a = "bundle::a"
    bundle_b = "bundle::b"
    nodes = {
        fn_key: {"kind": "fn", "label": "f"},
        bundle_a: {"kind": "bundle", "label": "a,b"},
        bundle_b: {"kind": "bundle", "label": "a,b"},
    }
    bundle_map = {bundle_a: {"a", "b"}, bundle_b: {"a", "b"}}
    adj = {fn_key: {bundle_a, bundle_b}}
    component = [fn_key, bundle_a, bundle_b]
    config_bundles_by_path = {tmp_path / "other.py": {"Cfg": {"x", "y"}}}
    documented_bundles_by_path = {path: {("a", "b")}}
    _, summary = da._render_mermaid_component(
        nodes,
        bundle_map,
        adj,
        component,
        config_bundles_by_path,
        documented_bundles_by_path,
    )
    assert "Declared Config bundles: none found for this component." in summary
    assert "Documented bundles" in summary
    assert "(tier-2, documented)" in summary


def test_merge_counts_by_knobs_merges_subset() -> None:
    da = _load()
    counts = {("a", "b"): 1, ("a", "b", "k"): 2}
    merged = da._merge_counts_by_knobs(counts, {"k"})
    assert merged == {("a", "b", "k"): 3}



from __future__ import annotations

from pathlib import Path
import sys
import textwrap


def _load():
    repo_root = Path(__file__).resolve().parents[1]
    sys.path.insert(0, str(repo_root / "src"))
    from gabion.analysis.dataflow_audit import analyze_paths, AuditConfig

    return analyze_paths, AuditConfig


def test_tuple_unpacking_preserves_bundle(tmp_path: Path) -> None:
    analyze_paths, AuditConfig = _load()
    source = textwrap.dedent(
        """
        def sink(x=None, y=None):
            return x, y

        def f(a, b):
            x, y = a, b
            sink(x=x, y=x)
            sink(x=y, y=y)
        """
    ).lstrip()
    file_path = tmp_path / "mod.py"
    file_path.write_text(source)
    config = AuditConfig(project_root=tmp_path)
    analysis = analyze_paths(
        [file_path],
        recursive=True,
        type_audit=False,
        type_audit_report=False,
        type_audit_max=0,
        include_constant_smells=False,
        include_unused_arg_smells=False,
        config=config,
    )
    bundles = []
    for groups in analysis.groups_by_path.values():
        for group_list in groups.values():
            bundles.extend(group_list)
    assert any(bundle == {"a", "b"} for bundle in bundles)
